from __future__ import annotations

import abc
import socket
import typing
from binascii import hexlify
from struct import Struct, pack, unpack_from

from .interfaces.udp.endpoint import DomainAddress, UDPv4Address, UDPv6Address

ADDRESS_TYPE_IPV4 = 0x01
ADDRESS_TYPE_DOMAIN_NAME = 0x02
ADDRESS_TYPE_IPV6 = 0x03


FormatListType = typing.Union[str, "Serializable", typing.List["FormatListType"]]  # type:ignore


class PackError(RuntimeError):
    pass


class NestedPayload(object):
    """
    This is a special type of format. Allowing for nested packing.

    You can specify which serializable to use by specifying its class in the format_list of the parent Serializable.

    For example, nesting a Serializable of class B in class A::

        class A(Serializable):
            format_list = [B]

        def __init__(self, b_instance):
            pass

        def to_pack_list(self):
            return [("payload", B())]

        @classmethod
        def from_unpack_list(cls, *args):
            return A(*args)

    """

    def __init__(self, serializer):
        """
        Initialize with a known serializer, so we do not keep creating Serializer instances.
        As an added bonus, we also get all of the custom types defined in the given instance.

        :param serializer: the Serializer to inherit
        :type serializer: Serializer
        """
        super(NestedPayload, self).__init__()
        self.serializer = serializer

    def pack(self, serializable):
        """
        Pack some serializable.

        :param serializable: the Serializable instance which we should serialize.
        :type serializable: Serializable
        :return: the serialized data
        :rtype: bytes
        """
        data = self.serializer.pack_serializable(serializable)
        data = pack('>H', len(data)) + data
        return data

    def unpack(self, data, offset, unpack_list, serializable_class):
        """
        Unpack a Serializable using a class definition for some given data and offset.
        This is a special unpack_from which also takes a payload class.

        :param data: the data to unpack from
        :type data: str
        :param offset: the offset in the list of data to unpack from
        :type offset: int
        :param unpack_list: the list to which to append the Serializable
        :type unpack_list: list
        :param serializable_class: the Serializable class to unpack to
        :type serializable_class: type(Serializable)
        :return: the new offset
        :rtype: int
        """
        unpacked, offset = self.serializer.unpack_serializable(serializable_class, data, offset=offset + 2)
        unpack_list.append(unpacked)
        return offset


class Bits(object):

    def pack(self, *data):
        """
        Pack multiple bits into a single byte.

        :param *data: bit values
        :type *data: list of 8 True or False values (or anything that maps to it in an if-statement)
        """
        byte = 0
        byte |= 0x80 if data[0] else 0x00
        byte |= 0x40 if data[1] else 0x00
        byte |= 0x20 if data[2] else 0x00
        byte |= 0x10 if data[3] else 0x00
        byte |= 0x08 if data[4] else 0x00
        byte |= 0x04 if data[5] else 0x00
        byte |= 0x02 if data[6] else 0x00
        byte |= 0x01 if data[7] else 0x00
        return pack('>B', byte)

    def unpack(self, data, offset, unpack_list):
        """
        Unpack multiple bits from a single byte. The resulting bits are appended to unpack_list

        :returns: the new offset
        """
        byte, = unpack_from('>B', data, offset)
        bit_7 = 1 if 0x80 & byte else 0
        bit_6 = 1 if 0x40 & byte else 0
        bit_5 = 1 if 0x20 & byte else 0
        bit_4 = 1 if 0x10 & byte else 0
        bit_3 = 1 if 0x08 & byte else 0
        bit_2 = 1 if 0x04 & byte else 0
        bit_1 = 1 if 0x02 & byte else 0
        bit_0 = 1 if 0x01 & byte else 0
        unpack_list += [bit_7, bit_6, bit_5, bit_4, bit_3, bit_2, bit_1, bit_0]
        return offset + 1


class Raw(object):
    """
    Paste/unpack the remaining input without (un)packing.
    """

    def pack(self, packable):
        return packable

    def unpack(self, data, offset, unpack_list):
        unpack_list.append(data[offset:])
        return len(data)


class VarLen(object):
    """
    Pack/unpack from an encoded length + data string.
    """

    def __init__(self, length_format, base=1):
        self.length_format = length_format
        self.length_size = Struct(length_format).size
        self.base = base

    def pack(self, data):
        return pack(self.length_format, len(data) // self.base) + data

    def unpack(self, data, offset, unpack_list):
        str_length = unpack_from(self.length_format, data, offset)[0] * self.base
        unpack_list.append(data[offset + self.length_size: offset + self.length_size + str_length])
        return offset + self.length_size + str_length


class IPv4:
    """
    Pack/unpack an IPv4 address
    """

    def pack(self, data):
        return pack('>4sH', socket.inet_aton(data[0]), data[1])

    def unpack(self, data, offset, unpack_list):
        host_bytes, port = unpack_from('>4sH', data, offset)
        unpack_list.append(UDPv4Address(socket.inet_ntoa(host_bytes), port))
        return offset + 6


class Address:

    def __init__(self, ip_only=False):
        self.ip_only = ip_only

    def pack(self, address):
        if isinstance(address, UDPv6Address):
            return pack('>B16sH', ADDRESS_TYPE_IPV6, socket.inet_pton(socket.AF_INET6, address.ip), address.port)
        if not self.ip_only and isinstance(address, DomainAddress):
            host_bytes = address.host.encode()
            return pack(f'>BH{len(host_bytes)}sH', ADDRESS_TYPE_DOMAIN_NAME, len(host_bytes), host_bytes, address.port)
        if isinstance(address, tuple):
            return pack('>B4sH', ADDRESS_TYPE_IPV4, socket.inet_pton(socket.AF_INET, address[0]), address[1])
        raise PackError(f'Unexpected address type {address}')

    def unpack(self, data, offset, unpack_list):
        address_type, = unpack_from('>B', data, offset)
        if address_type == ADDRESS_TYPE_IPV4:
            ip_bytes, port = unpack_from('>4sH', data, offset + 1)
            unpack_list.append(UDPv4Address(socket.inet_ntop(socket.AF_INET, ip_bytes), port))
            return offset + 7
        elif address_type == ADDRESS_TYPE_IPV6:
            ip_bytes, port = unpack_from('>16sH', data, offset + 1)
            unpack_list.append(UDPv6Address(socket.inet_ntop(socket.AF_INET6, ip_bytes), port))
            return offset + 19
        elif not self.ip_only and address_type == ADDRESS_TYPE_DOMAIN_NAME:
            length, = unpack_from('>H', data, offset + 1)
            host = data[offset + 3: offset + 3 + length].decode()
            unpack_list.append(DomainAddress(host, unpack_from('>H', data, offset + 3 + length)[0]))
            return offset + 5 + length
        else:
            raise PackError(f'Cannot unpack address type {address_type}')


class ListOf:

    def __init__(self, packer, length_format='>B'):
        self.packer = packer
        self.length_format = length_format
        self.length_size = Struct(length_format).size

    def pack(self, data):
        return pack(self.length_format, len(data)) + b''.join([self.packer.pack(item) for item in data])

    def unpack(self, data, offset, unpack_list, *args):
        length, = unpack_from(self.length_format, data, offset)
        offset += self.length_size

        result = []
        unpack_list.append(result)
        for _ in range(length):
            offset = self.packer.unpack(data, offset, result, *args)
        return offset


class DefaultStruct:

    def __init__(self, format_str):
        self.format_str = format_str
        self.size = Struct(format_str).size

    def pack(self, *data):
        return pack(self.format_str, *data)

    def unpack(self, data, offset, unpack_list):
        result = unpack_from(self.format_str, data, offset)
        unpack_list.append(result if len(result) > 1 else result[0])
        return offset + self.size


class Serializer(object):

    def __init__(self):
        super(Serializer, self).__init__()
        self._packers = {
            '?': DefaultStruct(">?"),
            'B': DefaultStruct(">B"),
            'BBH': DefaultStruct(">BBH"),
            'BH': DefaultStruct(">BH"),
            'c': DefaultStruct(">c"),
            'f': DefaultStruct(">f"),
            'd': DefaultStruct(">d"),
            'H': DefaultStruct(">H"),
            'HH': DefaultStruct(">HH"),
            'I': DefaultStruct(">I"),
            'l': DefaultStruct(">l"),
            'LL': DefaultStruct(">LL"),
            'q': DefaultStruct(">q"),
            'Q': DefaultStruct(">Q"),
            'QH': DefaultStruct(">QH"),
            'QL': DefaultStruct(">QL"),
            'QQHHBH': DefaultStruct(">QQHHBH"),
            'ccB': DefaultStruct(">ccB"),
            '4SH': DefaultStruct(">4sH"),
            '20s': DefaultStruct(">20s"),
            '32s': DefaultStruct(">32s"),
            '64s': DefaultStruct(">64s"),
            '74s': DefaultStruct(">74s"),
            'c20s': DefaultStruct(">c20s"),
            'bits': Bits(),
            'ipv4': IPv4(),
            'ip_address': Address(ip_only=True),
            'address': Address(),
            'raw': Raw(),
            'varlenBx2': VarLen('>B', 2),
            'varlenH': VarLen('>H'),
            'varlenHx20': VarLen('>H', 20),
            'varlenH-list': ListOf(VarLen('>H')),
            'varlenI': VarLen('>I'),
            'doublevarlenH': VarLen('>H'),
            'payload': NestedPayload(self),
            'payload-list': ListOf(NestedPayload(self))
        }

    def get_available_formats(self):
        """
        Get all available packing formats.
        """
        return list(self._packers.keys())

    def get_packer_for(self, name):
        """
        Get a packer by name.
        """
        return self._packers[name]

    def add_packer(self, name, packer):
        """
        Register a new packer with a certain name.

        :param name: the name to register
        :param packer: the packer to use for it
        """
        self._packers[name] = packer

    def pack(self, fmt, item):
        """
        Pack data without using a Serializable. Using a Serializable is the preferred method.
        :param fmt: the name of the packer to use while packing
        :param item: object to pack
        :return: the packed data
        :rtype: bytes
        """
        return self._packers[fmt].pack(item)

    def unpack(self, fmt, data, offset=0):
        """
        Unpack data without using a Serializable. Using a Serializable is the preferred method.
        :param fmt: the name of the packer to use while unpacking
        :param data: bytes to unpack
        :return: the unpacked object
        :rtype: object
        """
        unpack_list = []
        if isinstance(fmt, str):
            offset = self._packers[fmt].unpack(data, offset, unpack_list)
        elif isinstance(fmt, list):
            offset = self._packers['payload-list'].unpack(data, offset, unpack_list, fmt[0])
        else:
            offset = self._packers['payload'].unpack(data, offset, unpack_list, fmt)
        return unpack_list[0], offset

    def pack_serializable(self, serializable):
        """
        Serialize a single Serializable instance.

        :param serializable: the Serializable to pack
        :type serializable: Serializable
        :return: the serialized object
        :rtype: bytes
        """
        packed = b''
        for packable in serializable.to_pack_list():
            try:
                packed += self._packers[packable[0]].pack(*packable[1:])
            except Exception as e:
                raise PackError("Could not pack item: %s\n%s: %s" % (packable, type(e).__name__, str(e))) from e
        return packed

    def pack_serializable_list(self, serializables):
        """
        Serialize a list of Serializable instances.

        :param serializables: the Serializables to pack
        :type serializables: [Serializable]
        :return: the serialized list
        :rtype: bytes
        """
        return b''.join(self.pack_serializable(serializable) for serializable in serializables)

    def unpack_serializable(self, serializable, data, offset=0):
        """
        Use the formats specified in a serializable object and unpack to it.

        :param serializable: the serializable classes to get the format from and unpack to
        :param data: the data to unpack from
        :param offset: the optional offset to unpack data from
        """
        unpack_list = []
        for fmt in serializable.format_list:
            try:
                offset = self._packers[fmt].unpack(data, offset, unpack_list)
            except KeyError:
                if not issubclass(fmt, Serializable):
                    raise
                offset = self._packers['payload'].unpack(data, offset, unpack_list, fmt)
            except TypeError:
                if not isinstance(fmt, list):
                    raise
                offset = self._packers['payload-list'].unpack(data, offset, unpack_list, fmt[0])
            except Exception as e:
                raise PackError("Could not unpack item: %s\n%s: %s" % (fmt, type(e).__name__, str(e))) from e
        return serializable.from_unpack_list(*unpack_list), offset

    def unpack_serializable_list(self, serializables, data, offset=0, consume_all=True):
        """
        Use the formats specified in a list of serializable objects and unpack to them.

        :param serializables: the serializable classes to get the format from and unpack to
        :param data: the data to unpack from
        :param offset: position at which to start reading from data
        :param consume_all: if having a non-empty remainder should throw an error
        :except PackError: if the data could not be fit into the specified serializables
        :except PackError: if consume_all is True and not all of the data was consumed when parsing the serializables
        :return: the list of Serializable instances
        :rtype: [Serializable]
        """
        unpacked = []
        for serializable in serializables:
            payload, offset = self.unpack_serializable(serializable, data, offset)
            unpacked.append(payload)
        remainder = data[offset:]
        if not consume_all:
            unpacked.append(remainder)
        elif remainder:
            raise PackError("Incoming packet %s (%s) has extra data: (%s)" %
                            (str([serializable_class.__name__ for serializable_class in serializables]),
                             hexlify(data),
                             hexlify(remainder)))
        return unpacked


class Serializable(metaclass=abc.ABCMeta):
    """
    Interface for serializable objects.
    """

    format_list: typing.List[FormatListType] = []

    @abc.abstractmethod
    def to_pack_list(self) -> typing.List[tuple]:
        """
        Serialize this object to a Serializer pack list.

        E.g.:
        ``[(format1, data1), (format2, data2), (format3, data3), ..]``
        """
        pass

    @abc.abstractmethod
    def from_unpack_list(cls, *args) -> Serializable:  # pylint: disable=E0213
        """
        Create a new Serializable object from a list of unpacked variables.
        """
        pass


class Payload(Serializable, abc.ABC):

    def __str__(self):
        out = self.__class__.__name__
        for attribute in dir(self):
            if not (attribute.startswith('_') or callable(getattr(self, attribute))) \
                    and attribute not in ['format_list', 'names']:
                out += '\n| %s: %s' % (attribute, repr(getattr(self, attribute)))
        return out


# Serializers should be stateless.
# Therefore we can expose a global singleton for efficiency.
# If you do need a Serializer with a state, be sure to use your own instance.
default_serializer = Serializer()



from asyncio import get_event_loop
from collections import defaultdict

from ipv8.test.mocking.endpoint import AutoMockEndpoint
from ipv8.util import succeed


class SimulationEndpoint(AutoMockEndpoint):
    """
    Endpoint used in a simulated IPv8 environment. We make the open function async since this is expected by
    the IPv8 service.
    """

    def __init__(self):
        super().__init__()
        self.latencies = defaultdict(int)

    async def open(self):
        self._open = True
        return succeed(None)

    def get_link_latency(self, to_address):
        """
        Return the link latency to a particular destination address, in seconds.
        """
        return self.latencies[to_address]

    def send(self, socket_address, packet):
        get_event_loop().call_later(self.get_link_latency(socket_address), super().send,
                                    socket_address, packet)



import base64

from aiohttp import web

from aiohttp_apispec import docs, json_schema

from marshmallow.fields import Dict, Float, String

from .base_endpoint import BaseEndpoint, HTTP_BAD_REQUEST, Response
from .schema import DefaultResponseSchema, schema
from ..attestation.communication_manager import CommunicationManager
from ..util import strip_sha1_padding


PseudonymListResponseSchema = schema(PseudonymListResponse={"names": [String]})
CredentialSchema = schema(Credential={"name": String, "hash": String, "metadata": Dict, "attesters": [String]})
CredentialListResponseSchema = schema(CredentialListResponse={"names": [CredentialSchema]})


def ez_b64_encode(s):
    return base64.b64encode(s).decode()


def ez_b64_decode(s):
    return base64.b64decode(s.encode())


class IdentityEndpoint(BaseEndpoint):

    def __init__(self, middlewares=()):
        super(IdentityEndpoint, self).__init__(middlewares)
        self.communication_manager = None
        self.app.on_shutdown.append(self.on_shutdown)

    def initialize(self, session):
        super(IdentityEndpoint, self).initialize(session)
        self.communication_manager = CommunicationManager(session)

    def setup_routes(self):
        self.app.add_routes([web.get('', self.list_pseudonyms),

                             web.get('/{pseudonym_name}/schemas', self.list_schemas),

                             web.get('/{pseudonym_name}/public_key', self.get_pseudonym_public_key),
                             web.get('/{pseudonym_name}/unload', self.unload_pseudonym),
                             web.get('/{pseudonym_name}/remove', self.remove_pseudonym),

                             web.get('/{pseudonym_name}/credentials', self.list_pseudonym_credentials),
                             web.get('/{pseudonym_name}/credentials/{subject_key}', self.list_subject_credentials),
                             web.get('/{pseudonym_name}/peers', self.list_pseudonym_peers),

                             web.put('/{pseudonym_name}/allow/{verifier_key}', self.allow_pseudonym_verification),
                             web.put('/{pseudonym_name}/disallow/{verifier_key}', self.disallow_pseudonym_verification),
                             web.put('/{pseudonym_name}/request/{authority_key}', self.create_pseudonym_credential),

                             web.put('/{pseudonym_name}/attest/{subject_key}', self.attest_pseudonym_credential),
                             web.put('/{pseudonym_name}/verify/{subject_key}', self.verify_pseudonym_credential),

                             web.get('/{pseudonym_name}/outstanding/attestations', self.list_pseudonym_outstanding_attestations),
                             web.get('/{pseudonym_name}/outstanding/verifications', self.list_pseudonym_outstanding_verifications),

                             web.get('/{pseudonym_name}/verifications', self.list_pseudonym_verification_output)
                             ])

    @docs(
        tags=["Identity"],
        summary="List our pseudonyms.",
        responses={
            200: {
                "schema": PseudonymListResponseSchema
            }
        }
    )
    async def list_pseudonyms(self, request):
        return Response({"names": self.communication_manager.list_names()})

    @docs(
        tags=["Identity"],
        summary="List our available identity schemas.",
        parameters=[{
            'in': 'path',
            'name': 'pseudonym_name',
            'description': 'The name of the pseudonym to use.',
            'type': 'string',
        }],
        responses={
            200: {
                "schema": schema(SchemaListResponse={"schemas": [String]})
            }
        }
    )
    async def list_schemas(self, request):
        channel = await self.communication_manager.load(request.match_info['pseudonym_name'],
                                                        request.headers.get('X-Rendezvous'))
        return Response({"schemas": channel.schemas})

    @docs(
        tags=["Identity"],
        summary="Get the public key for a pseudonym.",
        parameters=[{
            'in': 'path',
            'name': 'pseudonym_name',
            'description': 'The name of the pseudonym to use.',
            'type': 'string',
        }],
        responses={
            200: {
                "schema": PseudonymListResponseSchema
            }
        }
    )
    async def get_pseudonym_public_key(self, request):
        channel = await self.communication_manager.load(request.match_info['pseudonym_name'],
                                                        request.headers.get('X-Rendezvous'))
        return Response({"public_key": ez_b64_encode(channel.public_key_bin)})

    @docs(
        tags=["Identity"],
        summary="Unload a pseudonym.",
        parameters=[{
            'in': 'path',
            'name': 'pseudonym_name',
            'description': 'The name of the pseudonym to use.',
            'type': 'string',
        }],
        responses={
            200: {
                "schema": DefaultResponseSchema,
                "example": {"success": True}
            }
        }
    )
    async def unload_pseudonym(self, request):
        await self.communication_manager.unload(request.match_info['pseudonym_name'])
        return Response({"success": True})

    @docs(
        tags=["Identity"],
        summary="Remove a pseudonym.",
        parameters=[{
            'in': 'path',
            'name': 'pseudonym_name',
            'description': 'The name of the pseudonym to use.',
            'type': 'string',
        }],
        responses={
            200: {
                "schema": DefaultResponseSchema,
                "example": {"success": True}
            }
        }
    )
    async def remove_pseudonym(self, request):
        await self.communication_manager.remove(request.match_info['pseudonym_name'])
        return Response({"success": True})

    @docs(
        tags=["Identity"],
        summary="List a pseudonym's credentials.",
        parameters=[{
            'in': 'path',
            'name': 'pseudonym_name',
            'description': 'The name of the pseudonym to use.',
            'type': 'string',
        }],
        responses={
            200: {
                "schema": CredentialListResponseSchema
            }
        }
    )
    async def list_pseudonym_credentials(self, request):
        channel = await self.communication_manager.load(request.match_info['pseudonym_name'],
                                                        request.headers.get('X-Rendezvous'))
        return Response({"names": [{
            "name": data[0],
            "hash": ez_b64_encode(strip_sha1_padding(attribute_hash)),
            "metadata": data[1],
            "attesters": [ez_b64_encode(attester) for attester in data[2]]
        }
            for attribute_hash, data in channel.get_my_attributes().items()]
        })

    @docs(
        tags=["Identity"],
        summary="List a subject's credentials.",
        parameters=[
            {
                'in': 'path',
                'name': 'pseudonym_name',
                'description': 'The name of the pseudonym to use.',
                'type': 'string',
            },
            {
                'in': 'path',
                'name': 'subject_key',
                'description': 'The public key of the subject to attest for.',
                'type': 'string',
            }
        ],
        responses={
            200: {
                "schema": CredentialListResponseSchema
            },
            400: {
                "schema": DefaultResponseSchema,
                "examples": {'Subject not found': {"success": False, "error": "failed to find subject"}}
            }
        }
    )
    async def list_subject_credentials(self, request):
        channel = await self.communication_manager.load(request.match_info['pseudonym_name'],
                                                        request.headers.get('X-Rendezvous'))

        subject = None
        for peer in channel.peers:
            if peer.public_key.key_to_bin() == ez_b64_decode(request.match_info['subject_key']):
                subject = peer
                break
        if subject is None:
            return Response({"success": False, "error": "failed to find subject"})

        return Response({"names": [{
            "name": data[0],
            "hash": ez_b64_encode(strip_sha1_padding(attribute_hash)),
            "metadata": data[1],
            "attesters": [ez_b64_encode(attester) for attester in data[2]]
        }
            for attribute_hash, data in channel.get_attributes(subject).items()]
        })

    @docs(
        tags=["Identity"],
        summary="List a pseudonym's peers.",
        parameters=[{
            'in': 'path',
            'name': 'pseudonym_name',
            'description': 'The name of the pseudonym to use.',
            'type': 'string',
        }],
        responses={
            200: {
                "schema": schema(PeerListResponse={"peers": [String]})
            }
        }
    )
    async def list_pseudonym_peers(self, request):
        channel = await self.communication_manager.load(request.match_info['pseudonym_name'],
                                                        request.headers.get('X-Rendezvous'))
        return Response({"peers": [ez_b64_encode(peer.public_key.key_to_bin()) for peer in channel.peers]})

    @docs(
        tags=["Identity"],
        summary="Verify a credential.",
        parameters=[
            {
                'in': 'path',
                'name': 'pseudonym_name',
                'description': 'The name of the pseudonym to use.',
                'type': 'string',
            },
            {
                'in': 'path',
                'name': 'verifier_key',
                'description': 'The public key of the verifier.',
                'type': 'string',
            }
        ],
        responses={
            200: {
                "schema": DefaultResponseSchema,
                "example": {"success": True}
            },
            400: {
                "schema": DefaultResponseSchema,
                "examples": {'Authority not found': {"success": False, "error": "failed to find authority"}}
            }
        }
    )
    @json_schema(schema(AllowVerification={
        'name*': (String, 'The name of the attribute to allow verification of.')
    }))
    async def allow_pseudonym_verification(self, request):
        parameters = await request.json()
        if 'name' not in parameters:
            return Response({"error": "incorrect parameters"}, status=HTTP_BAD_REQUEST)

        channel = await self.communication_manager.load(request.match_info['pseudonym_name'],
                                                        request.headers.get('X-Rendezvous'))
        verifier = None
        for peer in channel.peers:
            if peer.public_key.key_to_bin() == ez_b64_decode(request.match_info['verifier_key']):
                verifier = peer
                break
        if verifier is None:
            return Response({"success": False, "error": "failed to find verifier"})

        channel.allow_verification(verifier, parameters['name'])

        return Response({"success": True})

    @docs(
        tags=["Identity"],
        summary="Disallow verification of a credential.",
        parameters=[
            {
                'in': 'path',
                'name': 'pseudonym_name',
                'description': 'The name of the pseudonym to use.',
                'type': 'string',
            },
            {
                'in': 'path',
                'name': 'verifier_key',
                'description': 'The public key of the verifier.',
                'type': 'string',
            }
        ],
        responses={
            200: {
                "schema": DefaultResponseSchema,
                "example": {"success": True}
            },
            400: {
                "schema": DefaultResponseSchema,
                "examples": {'Authority not found': {"success": False, "error": "failed to find authority"}}
            }
        }
    )
    @json_schema(schema(DisallowVerification={
        'name*': (String, 'The name of the attribute to disallow verification of.')
    }))
    async def disallow_pseudonym_verification(self, request):
        parameters = await request.json()
        if 'name' not in parameters:
            return Response({"error": "incorrect parameters"}, status=HTTP_BAD_REQUEST)

        channel = await self.communication_manager.load(request.match_info['pseudonym_name'],
                                                        request.headers.get('X-Rendezvous'))
        verifier = None
        for peer in channel.peers:
            if peer.public_key.key_to_bin() == ez_b64_decode(request.match_info['verifier_key']):
                verifier = peer
                break
        if verifier is None:
            return Response({"success": False, "error": "failed to find verifier"})

        channel.disallow_verification(verifier, parameters['name'])

        return Response({"success": True})

    @docs(
        tags=["Identity"],
        summary="Create a credential.",
        parameters=[
            {
                'in': 'path',
                'name': 'pseudonym_name',
                'description': 'The name of the pseudonym to use.',
                'type': 'string',
            },
            {
                'in': 'path',
                'name': 'authority_key',
                'description': 'The public key of the authority to request attestation from.',
                'type': 'string',
            }
        ],
        responses={
            200: {
                "schema": DefaultResponseSchema,
                "example": {"success": True}
            },
            400: {
                "schema": DefaultResponseSchema,
                "examples": {'Authority not found': {"success": False, "error": "failed to find authority"}}
            }
        }
    )
    @json_schema(schema(AttestationRequest={
        'name*': (String, 'The name of the attribute to request attestation for.'),
        'schema*': (String, 'The attribute schema to use.'),
        'metadata': (Dict, 'The metadata to attach.')
    }))
    async def create_pseudonym_credential(self, request):
        parameters = await request.json()
        if 'name' not in parameters or 'schema' not in parameters:
            return Response({"error": "incorrect parameters"}, status=HTTP_BAD_REQUEST)

        channel = await self.communication_manager.load(request.match_info['pseudonym_name'],
                                                        request.headers.get('X-Rendezvous'))
        authority = None
        for peer in channel.peers:
            if peer.public_key.key_to_bin() == ez_b64_decode(request.match_info['authority_key']):
                authority = peer
                break
        if authority is None:
            return Response({"success": False, "error": "failed to find authority"})

        metadata = parameters['metadata'] if 'metadata' in parameters else {}
        channel.request_attestation(authority, parameters['name'], parameters['schema'], metadata)

        return Response({"success": True})

    @docs(
        tags=["Identity"],
        summary="Attest to a credential.",
        parameters=[
            {
                'in': 'path',
                'name': 'pseudonym_name',
                'description': 'The name of the pseudonym to use.',
                'type': 'string',
            },
            {
                'in': 'path',
                'name': 'subject_key',
                'description': 'The public key of the subject to attest for.',
                'type': 'string',
            }
        ],
        responses={
            200: {
                "schema": DefaultResponseSchema,
                "example": {"success": True}
            },
            400: {
                "schema": DefaultResponseSchema,
                "examples": {'Subject not found': {"success": False, "error": "failed to find subject"}}
            }
        }
    )
    @json_schema(schema(Attestation={
        'name*': (String, 'The name of the subject\'s attribute.'),
        'value*': (String, 'The value we believe the subject\'s attribute has.')
    }))
    async def attest_pseudonym_credential(self, request):
        parameters = await request.json()
        if 'name' not in parameters or 'value' not in parameters:
            return Response({"error": "incorrect parameters"}, status=HTTP_BAD_REQUEST)

        channel = await self.communication_manager.load(request.match_info['pseudonym_name'],
                                                        request.headers.get('X-Rendezvous'))

        subject = None
        for peer in channel.peers:
            if peer.public_key.key_to_bin() == ez_b64_decode(request.match_info['subject_key']):
                subject = peer
                break
        if subject is None:
            return Response({"success": False, "error": "failed to find subject"})

        channel.attest(subject, parameters["name"], ez_b64_decode(parameters["value"]))

        return Response({"success": True})

    @docs(
        tags=["Identity"],
        summary="Request verification of a credential.",
        parameters=[
            {
                'in': 'path',
                'name': 'pseudonym_name',
                'description': 'The name of the pseudonym to use.',
                'type': 'string',
            },
            {
                'in': 'path',
                'name': 'subject_key',
                'description': 'The public key of the subject to verify.',
                'type': 'string',
            }
        ],
        responses={
            200: {
                "schema": DefaultResponseSchema,
                "example": {"success": True}
            },
            400: {
                "schema": DefaultResponseSchema,
                "examples": {'Subject not found': {"success": False, "error": "failed to find subject"}}
            }
        }
    )
    @json_schema(schema(VerificationRequest={
        'hash*': (String, 'The hash of the subject\'s attribute.'),
        'value*': (String, 'The value we require the subject\'s attribute to have.'),
        'schema*': (String, 'The schema we require the subject\'s attribute to have.')
    }))
    async def verify_pseudonym_credential(self, request):
        parameters = await request.json()
        if 'hash' not in parameters or 'value' not in parameters or 'schema' not in parameters:
            return Response({"error": "incorrect parameters"}, status=HTTP_BAD_REQUEST)

        channel = await self.communication_manager.load(request.match_info['pseudonym_name'],
                                                        request.headers.get('X-Rendezvous'))

        subject = None
        for peer in channel.peers:
            if peer.public_key.key_to_bin() == ez_b64_decode(request.match_info['subject_key']):
                subject = peer
                break
        if subject is None:
            return Response({"success": False, "error": "failed to find subject"})

        channel.verify(subject, ez_b64_decode(parameters["hash"]), [ez_b64_decode(parameters["value"])],
                       parameters["schema"])

        return Response({"success": True})

    @docs(
        tags=["Identity"],
        summary="List the oustanding requests for attestations by others.",
        parameters=[{
            'in': 'path',
            'name': 'pseudonym_name',
            'description': 'The name of the pseudonym to use.',
            'type': 'string',
        }],
        responses={
            200: {
                "schema": schema(AttestationRequestsResponse={"requests": [schema(OutstandingAttestationRequest={
                    "peer": String,
                    "attribute_name": String,
                    "metadata": Dict
                })]})
            }
        }
    )
    async def list_pseudonym_outstanding_attestations(self, request):
        channel = await self.communication_manager.load(request.match_info['pseudonym_name'],
                                                        request.headers.get('X-Rendezvous'))
        formatted = []
        for k, v in channel.attestation_requests.items():
            formatted.append({
                "peer": ez_b64_encode(k[0].public_key.key_to_bin()),
                "attribute_name": k[1],
                "metadata": v[1]
            })
        return Response({"requests": formatted})

    @docs(
        tags=["Identity"],
        summary="List the oustanding requests for verification by others.",
        parameters=[{
            'in': 'path',
            'name': 'pseudonym_name',
            'description': 'The name of the pseudonym to use.',
            'type': 'string',
        }],
        responses={
            200: {
                "schema": schema(VerificationRequestsResponse={"requests": [schema(OutstandingVerificationRequest={
                    "peer": String,
                    "attribute_name": String
                })]})
            }
        }
    )
    async def list_pseudonym_outstanding_verifications(self, request):
        channel = await self.communication_manager.load(request.match_info['pseudonym_name'],
                                                        request.headers.get('X-Rendezvous'))
        formatted = []
        for k in channel.verify_requests.keys():
            formatted.append({
                "peer": ez_b64_encode(k[0].public_key.key_to_bin()),
                "attribute_name": k[1]
            })
        return Response({"requests": formatted})

    @docs(
        tags=["Identity"],
        summary="Return the output of our verification requests.",
        parameters=[
            {
                'in': 'path',
                'name': 'pseudonym_name',
                'description': 'The name of the pseudonym to use.',
                'type': 'string',
            }
        ],
        responses={
            200: {
                "schema": schema(VerificationOutputResponse={"outputs": [schema(VerificationOutput={
                    "hash": String,
                    "reference": String,
                    "match": Float
                })]})
            }
        }
    )
    async def list_pseudonym_verification_output(self, request):
        channel = await self.communication_manager.load(request.match_info['pseudonym_name'],
                                                        request.headers.get('X-Rendezvous'))

        formatted = []
        for k, v in channel.verification_output.items():
            if not v or v[0][1] is None:
                # Not done yet
                continue
            formatted.append({"hash": ez_b64_encode(k), "reference": ez_b64_encode(v[0][0]), "match": v[0][1]})

        return Response({"outputs": formatted})

    async def on_shutdown(self, _):
        await self.communication_manager.shutdown()



import time
from binascii import hexlify
from collections import defaultdict

from . import DHTError
from .community import DHTCommunity, MAX_NODES_IN_FIND, PING_INTERVAL, Request, TARGET_NODES, gather_without_errors
from .payload import (ConnectPeerRequestPayload, ConnectPeerResponsePayload, PingRequestPayload, PingResponsePayload,
                      StorePeerRequestPayload, StorePeerResponsePayload)
from .routing import NODE_STATUS_BAD, Node
from ..lazy_community import lazy_wrapper, lazy_wrapper_wd


class DHTDiscoveryCommunity(DHTCommunity):
    """
    Community for discovering peers that are behind NAT.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.store = defaultdict(list)
        self.store_for_me = defaultdict(list)

        self.add_message_handler(StorePeerRequestPayload, self.on_store_peer_request)
        self.add_message_handler(StorePeerResponsePayload, self.on_store_peer_response)
        self.add_message_handler(ConnectPeerRequestPayload, self.on_connect_peer_request)
        self.add_message_handler(ConnectPeerResponsePayload, self.on_connect_peer_response)

        self.register_task('store_peer', self.store_peer, interval=30)
        self.register_task('ping_all', self.ping_all, interval=10)

    @lazy_wrapper_wd(PingRequestPayload)
    def on_ping_request(self, peer, payload, data):
        super(DHTDiscoveryCommunity, self).on_ping_request(peer.address, data)
        node = self.find_node_in_dict(peer.key.key_to_bin(), self.store)
        if node:
            node.last_queries.append(time.time())

    @lazy_wrapper_wd(PingResponsePayload)
    def on_ping_response(self, peer, payload, data):
        super(DHTDiscoveryCommunity, self).on_ping_response(peer.address, data)
        node = self.find_node_in_dict(peer.key.key_to_bin(), self.store_for_me)
        if node:
            node.last_response = time.time()

    def find_node_in_dict(self, public_key_bin, node_dict):
        for _, nodes in node_dict.items():
            for node in nodes:
                if node.public_key.key_to_bin() == public_key_bin:
                    return node

    async def store_peer(self):
        key = self.my_peer.mid

        # Do we already have enough peers storing our address?
        if len(self.store_for_me[key]) >= TARGET_NODES // 2:
            return []

        try:
            nodes = await self.find_nodes(key)
            return await self.send_store_peer_request(key, nodes[:TARGET_NODES])
        except DHTError:
            return []

    async def send_store_peer_request(self, key, nodes):
        # Create new objects to avoid problem with the nodes also being in the routing table
        nodes = [Node(node.key, node.address) for node in nodes if node not in self.store_for_me[key]]

        if not nodes:
            return DHTError('No nodes found for storing peer')

        futures = []
        for node in nodes:
            if node.id in self.tokens:
                cache = Request(self, 'store-peer', node, [key])
                self.request_cache.add(cache)
                futures.append(cache.future)
                self.ez_send(node, StorePeerRequestPayload(cache.number, self.tokens[node.id][1], key))
            else:
                self.logger.debug('Not sending store-peer-request to %s (no token available)', node)

        if not futures:
            raise DHTError('Peer was not stored')
        return await gather_without_errors(*futures)

    async def connect_peer(self, mid, peer=None):
        if mid in self.store:
            return self.store[mid]

        # If a peer is provided, we will first try to ping the peer (to see if it's connectable).
        # This could potentially save an expensive DHT lookup.
        if peer:
            node = Node(peer.key, peer.address)
            try:
                await self.ping(node)
            except DHTError:
                pass
            else:
                return [node]

        nodes = await self.find_nodes(mid)
        nodes = await self.send_connect_peer_request(mid, nodes[:TARGET_NODES])
        if not nodes:
            raise DHTError('Failed to connect peer')
        return nodes

    async def send_connect_peer_request(self, key, nodes):
        # Create new objects to avoid problem with the nodes also being in the routing table
        nodes = [Node(node.key, node.address) for node in nodes]

        if not nodes:
            raise DHTError('No nodes found for connecting to peer')

        futures = []
        for node in nodes:
            cache = Request(self, 'connect-peer', node)
            self.request_cache.add(cache)
            futures.append(cache.future)
            self.ez_send(node, ConnectPeerRequestPayload(cache.number, self.my_estimated_lan, key))

        node_lists = await gather_without_errors(*futures)
        return list(set(sum(node_lists, [])))

    @lazy_wrapper(StorePeerRequestPayload)
    def on_store_peer_request(self, peer, payload):
        self.logger.debug('Got store-peer-request from %s', peer.address)

        node = Node(peer.key, peer.address)
        node.last_queries.append(time.time())

        if not self.check_token(node, payload.token):
            self.logger.warning('Bad token, dropping packet.')
            return
        if payload.target != peer.mid:
            self.logger.warning('Not allowed to store under key %s, dropping packet.', hexlify(payload.target))
            return

        if node not in self.store[payload.target]:
            self.logger.debug('Storing peer %s (key %s)', node, hexlify(payload.target))
            self.store[payload.target].append(node)

        self.ez_send(node, StorePeerResponsePayload(payload.identifier))

    @lazy_wrapper(StorePeerResponsePayload)
    def on_store_peer_response(self, peer, payload):
        if not self.request_cache.has('store-peer', payload.identifier):
            self.logger.error('Got store-peer-response with unknown identifier, dropping packet')
            return

        self.logger.debug('Got store-peer-response from %s', peer.address)

        cache = self.request_cache.pop('store-peer', payload.identifier)

        key = cache.params[0]
        if cache.node not in self.store_for_me[key]:
            self.logger.debug('Peer %s storing us (key %s)', cache.node, hexlify(key))
            self.store_for_me[key].append(cache.node)

        cache.future.set_result(cache.node)

    @lazy_wrapper(ConnectPeerRequestPayload)
    def on_connect_peer_request(self, peer, payload):
        self.logger.debug('Got connect-peer-request from %s', peer.address)

        nodes = self.store[payload.target][:MAX_NODES_IN_FIND]
        for node in nodes:
            packet = self.create_puncture_request(payload.lan_address, peer.address, payload.identifier)
            self.endpoint.send(node.address, packet)

        self.logger.debug('Returning peers %s (key %s)', nodes, hexlify(payload.target))
        self.ez_send(peer, ConnectPeerResponsePayload(payload.identifier, nodes))

    @lazy_wrapper(ConnectPeerResponsePayload)
    def on_connect_peer_response(self, peer, payload):
        if not self.request_cache.has('connect-peer', payload.identifier):
            self.logger.error('Got connect-peer-response with unknown identifier, dropping packet')
            return

        self.logger.debug('Got connect-peer-response from %s', peer.address)
        cache = self.request_cache.pop('connect-peer', payload.identifier)
        cache.future.set_result(payload.nodes)

    def ping_all(self):
        now = time.time()
        for key, nodes in self.store_for_me.items():
            for index in range(len(nodes) - 1, -1, -1):
                node = nodes[index]
                if node.status == NODE_STATUS_BAD:
                    self.store_for_me[key].pop(index)
                    self.logger.debug('Deleting peer %s that stored us (key %s)', node, hexlify(key))
                elif node.last_ping_sent + PING_INTERVAL <= now:
                    self.ping(node)

        for key, nodes in self.store.items():
            for index in range(len(nodes) - 1, -1, -1):
                node = nodes[index]
                if now > node.last_query + 60:
                    self.logger.debug('Deleting peer %s (key %s)', node, hexlify(key))
                    self.store[key].pop(index)



from abc import ABC, abstractmethod
from functools import wraps
from typing import Iterable, List, Tuple, Type, Union

from .keyvault.crypto import default_eccrypto
from .messaging.payload_headers import BinMemberAuthenticationPayload, GlobalTimeDistributionPayload
from .overlay import Overlay
from .peer import Peer
from .types import Address, NumberCache, Payload


def cache_retrieval_failed(overlay: Overlay, cache_class: Type[NumberCache]) -> None:
    """
    Handler for messages which failed to match to an existing overlay cache.
    """
    overlay.logger.debug("Failed to match %s: was answered late or did not exist.", repr(cache_class))


def retrieve_cache(cache_class: Type[NumberCache]):
    """
    This function wrapper match a payload to a registered cache for you.

    For this wrapper to function, you will need to comply with three standards:

     - The last specified payload must include an ``identifier`` attribute.
     - The ``cache_class`` must specify a ``name`` attribute.
     - The overlay this method belongs to has a ``request_cache`` attribute.

    You can now message handlers as follows:

    ::

        @lazy_wrapper(MyPayload1, MyPayload2)  # MyPayload2.identifier must exist!
        @retrieve_cache(MyCache)  # MyCache.name must exist!
        def on_message(peer: Peer,
                       payload1: MyPayload1,
                       payload2: MyPayload2,
                       cache: NumberCache):
            pass

    :param cache_class: the cache to fetch.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, peer_or_addr, *payloads):
            try:
                # For the ``_wd`` decorators, the last argument is the data and not the Payload.
                payload = payloads[-2 if isinstance(payloads[-1], bytes) else -1]
                cache = self.request_cache.pop(cache_class.name, payload.identifier)
                return func(self, peer_or_addr, *payloads, cache=cache)
            except KeyError:
                return cache_retrieval_failed(self, cache_class)

        return wrapper

    return decorator


def lazy_wrapper(*payloads: Type[Payload]):
    """
    This function wrapper will unpack the BinMemberAuthenticationPayload for you.

    You can now write your authenticated and signed functions as follows:

    ::

        @lazy_wrapper(IntroductionRequestPayload, IntroductionResponsePayload)
        def on_message(peer: Peer,
                       payload1: IntroductionRequestPayload,
                       payload2: IntroductionResponsePayload):
            pass
    """
    def decorator(func):
        @wraps(func)
        def wrapper(self, source_address, data):
            # UNPACK
            auth, _ = self.serializer.unpack_serializable(BinMemberAuthenticationPayload, data, offset=23)
            signature_valid, remainder = self._verify_signature(auth, data)
            unpacked = self.serializer.unpack_serializable_list(payloads, remainder, offset=23)
            # ASSERT
            if not signature_valid:
                raise PacketDecodingError("Incoming packet %s has an invalid signature" %
                                          str([payload_class.__name__ for payload_class in payloads]))
            # PRODUCE
            peer = (self.network.verified_by_public_key_bin.get(auth.public_key_bin)
                    or Peer(auth.public_key_bin, source_address))
            return func(self, peer, *unpacked)
        return wrapper
    return decorator


def lazy_wrapper_wd(*payloads: Type[Payload]):
    """
    This function wrapper will unpack the BinMemberAuthenticationPayload for you, as well as pass the raw data to the
    decorated function

    You can now write your authenticated and signed functions as follows:

    ::

        @lazy_wrapper_wd(IntroductionRequestPayload, IntroductionResponsePayload)
        def on_message(peer: Peer,
                       payload1: IntroductionRequestPayload,
                       payload2: IntroductionResponsePayload,
                       data: bytes):
            pass
    """
    def decorator(func):
        @wraps(func)
        def wrapper(self, source_address, data):
            # UNPACK
            auth, _ = self.serializer.unpack_serializable(BinMemberAuthenticationPayload, data, offset=23)
            signature_valid, remainder = self._verify_signature(auth, data)
            unpacked = self.serializer.unpack_serializable_list(payloads, remainder, offset=23)
            # ASSERT
            if not signature_valid:
                raise PacketDecodingError("Incoming packet %s has an invalid signature" %
                                          str([payload_class.__name__ for payload_class in payloads]))
            # PRODUCE
            output = unpacked + [data]
            peer = (self.network.verified_by_public_key_bin.get(auth.public_key_bin)
                    or Peer(auth.public_key_bin, source_address))
            return func(self, peer, *output)
        return wrapper
    return decorator


def lazy_wrapper_unsigned(*payloads: Type[Payload]):
    """
    This function wrapper will unpack just the normal payloads for you.

    You can now write your non-authenticated and signed functions as follows:

    ::

        @lazy_wrapper_unsigned(IntroductionRequestPayload, IntroductionResponsePayload)
        def on_message(source_address: Address,
                       payload1: IntroductionRequestPayload,
                       payload2: IntroductionResponsePayload):
            pass
    """
    def decorator(func):
        @wraps(func)
        def wrapper(self, source_address, data):
            # UNPACK
            unpacked = self.serializer.unpack_serializable_list(payloads, data, offset=23)
            return func(self, source_address, *unpacked)
        return wrapper
    return decorator


def lazy_wrapper_unsigned_wd(*payloads: Type[Payload]):
    """
    This function wrapper will unpack just the normal payloads for you, as well as pass the raw data to the decorated
    function

    You can now write your non-authenticated and signed functions as follows:

    ::

        @lazy_wrapper_unsigned_wd(IntroductionRequestPayload, IntroductionResponsePayload)
        def on_message(source_address: Address,
                       payload1: IntroductionRequestPayload,
                       payload2: IntroductionResponsePayload,
                       data: bytes):
            pass
    """
    def decorator(func):
        @wraps(func)
        def wrapper(self, source_address, data):

            @lazy_wrapper_unsigned(*payloads)
            def inner_wrapper(inner_self, inner_source_address, *pyls):
                combo = list(pyls) + [data]
                return func(inner_self, inner_source_address, *combo)

            return inner_wrapper(self, source_address, data)
        return wrapper
    return decorator


class EZPackOverlay(Overlay, ABC):

    @abstractmethod
    def get_prefix(self) -> bytes:
        pass

    def ez_send(self, peer: Peer, *payloads: Payload, **kwargs) -> None:
        """
        Send a Payload instance (with a defined `msg_id` field) to a peer.
        If you supply more than one Payload instance, the `msg_id` of the LAST instance will be used.

        :param peer: the peer to send to
        :param sig: whether or not to sign this message
        :type sig: bool
        :param payloads: the list of Payload instances to serialize
        """
        self._ez_senda(peer.address, *payloads, **kwargs)

    def _ez_senda(self, address: Address, *payloads: Payload, **kwargs) -> None:
        """
        Send a Payload instance to an address.

        You will probably not need this, try to use `ez_send` instead.

        :param address: the address to send to
        :param sig: whether or not to sign this message
        :type sig: bool
        :param payloads: the list of Payload instances to serialize
        """

        # We promise the typing system that the ``msg_id`` is defined for the last payload.
        # Strictly speaking we should introduce a ``LastPayloadWithMessageID`` type, but this is annoying to work with.
        self.endpoint.send(address, self.ezr_pack(payloads[-1].msg_id, *payloads, **kwargs))  # type:ignore

    def ezr_pack(self, msg_num: int, *payloads: Payload, **kwargs) -> bytes:
        """
        The easier way to pack your messages. Supply with the message number and the Payloads you want to serialize.
        Optionally you can choose to sign the message.

        :param msg_num: the message number to claim for this message
        :param sig: whether or not to sign this message
        :type sig: bool
        :param payloads: the list of Payload instances to serialize
        :return: the serialized message
        """
        sig = kwargs.get('sig', True)
        if sig:
            payloads = (BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin()),) + payloads
        return self._ez_pack(self.get_prefix(), msg_num, payloads, sig)

    def _ez_pack(self, prefix: bytes, msg_num: int, payloads: Iterable[Payload], sig: bool = True) -> bytes:
        packet = prefix + bytes([msg_num]) + self.serializer.pack_serializable_list(payloads)
        if sig:
            packet += default_eccrypto.create_signature(self.my_peer.key, packet)
        return packet

    def _verify_signature(self, auth: BinMemberAuthenticationPayload, data: bytes) -> Tuple[bool, bytes]:
        ec = default_eccrypto
        public_key = ec.key_from_public_bin(auth.public_key_bin)
        signature_length = ec.get_signature_length(public_key)
        remainder = data[2 + len(auth.public_key_bin):-signature_length]
        signature = data[-signature_length:]
        return ec.is_valid_signature(public_key, data[:-signature_length], signature), remainder

    def _ez_unpack_auth(self,
                        payload_class: Type[Payload],
                        data: bytes) -> Tuple[BinMemberAuthenticationPayload, GlobalTimeDistributionPayload, Payload]:
        # UNPACK
        auth, _ = self.serializer.unpack_serializable(BinMemberAuthenticationPayload, data, offset=23)
        signature_valid, remainder = self._verify_signature(auth, data)
        format = [GlobalTimeDistributionPayload, payload_class]
        unpacked = self.serializer.unpack_serializable_list(format, remainder, offset=23)
        # ASSERT
        if not signature_valid:
            raise PacketDecodingError("Incoming packet %s has an invalid signature" % payload_class.__name__)
        # PRODUCE
        return auth, unpacked[0], unpacked[1]

    def _ez_unpack_noauth(self,
                          payload_class: Type[Payload],
                          data: bytes,
                          global_time: bool = True) -> Union[List[Payload], Payload]:
        # UNPACK
        format = [GlobalTimeDistributionPayload, payload_class] if global_time else [payload_class]
        unpacked = self.serializer.unpack_serializable_list(format, data, offset=23)
        # PRODUCE
        return unpacked if global_time else unpacked[0]


class PacketDecodingError(RuntimeError):
    pass



import random
import time
from collections import deque

from ...community import Community
from ...messaging.anonymization.tunnel import IntroductionPoint, PEER_SOURCE_PEX
from ...peer import Peer

PEX_VERSION = 1


class PexCommunity(Community):
    def __init__(self, *args, **kwargs):
        infohash = kwargs.pop('info_hash')
        self.community_id = (int.from_bytes(infohash, 'big') + PEX_VERSION).to_bytes(20, 'big')
        self._prefix = b'\x00' + self.version + self.community_id
        super().__init__(*args, **kwargs)

        self.intro_points = deque(maxlen=20)
        self.intro_points_for = []

    def get_intro_points(self):
        """
        Get a list of the most recent introduction points that were discovered using PexCommunity.
        :return : list of IntroductionPoint objects
        """

        # Remove old introduction points
        now = time.time()
        while self.intro_points and self.intro_points[-1].last_seen + 300 < now:
            self.intro_points.pop()

        my_peer = Peer(self.my_peer.key, self.my_estimated_wan)
        return list(self.intro_points) + [IntroductionPoint(my_peer, seeder_pk, PEER_SOURCE_PEX)
                                          for seeder_pk in self.intro_points_for]

    def start_announce(self, seeder_pk):
        """
        Start announcing yourself as an introduction point for a certain seeder.
        :param seeder_pk: public key of the seeder (in binary format)
        """
        if seeder_pk not in self.intro_points_for:
            self.intro_points_for.append(seeder_pk)

    def stop_announce(self, seeder_pk):
        """
        Stop announcing yourself as an introduction point for a certain seeder.
        :param seeder_pk: public key of the seeder (in binary format)
        """
        if seeder_pk in self.intro_points_for:
            self.intro_points_for.remove(seeder_pk)

    @property
    def done(self):
        return not bool(self.intro_points_for)

    def process_extra_bytes(self, peer, extra_bytes):
        if not extra_bytes:
            return

        for seeder_pk in self.serializer.unpack('varlenH-list', extra_bytes)[0]:
            ip = IntroductionPoint(peer, seeder_pk, PEER_SOURCE_PEX)
            if ip in self.intro_points:
                # Remove first to put introduction point at front of the deque.
                self.intro_points.remove(ip)
            # Add new introduction point (with up-to-date last_seen)
            self.intro_points.appendleft(ip)

    def introduction_request_callback(self, peer, dist, payload):
        self.process_extra_bytes(peer, payload.extra_bytes)

    def introduction_response_callback(self, peer, dist, payload):
        self.process_extra_bytes(peer, payload.extra_bytes)

    def create_introduction_request(self, socket_address, extra_bytes=b'', new_style=False):
        return super().create_introduction_request(socket_address, self.get_seeder_pks(), new_style)

    def create_introduction_response(self, lan_socket_address, socket_address, identifier,
                                     introduction=None, extra_bytes=b'', prefix=None, new_style=False):
        return super().create_introduction_response(lan_socket_address, socket_address, identifier, introduction,
                                                    self.get_seeder_pks(), new_style=new_style)

    def send_ping(self, peer):
        self.send_introduction_request(peer)

    def get_seeder_pks(self):
        pks = random.sample(self.intro_points_for, min(len(self.intro_points_for), 10))
        return self.serializer.pack('varlenH-list', pks)



import binascii
import json
import logging
import struct
import typing

from .attestation import Attestation
from .database import Credential, IdentityDatabase
from .metadata import Metadata
from ..tokentree.tree import TokenTree
from ...keyvault.crypto import ECCrypto
from ...keyvault.keys import PrivateKey
from ...types import PublicKey, Token


class PseudonymManager(object):
    """
    Object to manage interactions with our own pseudonym or one belonging to someone else.

    Never create one yourself, use `IdentityManager.get_pseudonym()` instead!
    """

    def __init__(self, database: IdentityDatabase,
                 public_key: typing.Optional[PublicKey] = None,
                 private_key: typing.Optional[PrivateKey] = None):
        super(PseudonymManager, self).__init__()

        self.database = database

        self.tree = TokenTree(public_key=public_key, private_key=private_key)
        self.credentials = []

        logging.info(f"Loading public key {binascii.hexlify(self.public_key.key_to_hash()).decode()} from database")
        for token in self.database.get_tokens_for(self.public_key):
            self.tree.elements[token.get_hash()] = token
        self.credentials = self.database.get_credentials_for(self.public_key)

    @property
    def public_key(self) -> PublicKey:
        return self.tree.public_key

    def add_credential(self, token: Token, metadata: Metadata, attestations=set()) -> typing.Optional[Credential]:
        """
        Add a credential to this pseudonym.

        :returns: the added credential on success, otherwise None
        """
        # If this token belongs to this chain, insert it.
        # Note: if the given metadata is invalid, the token is still inserted!
        if self.tree.gather_token(token) is not None:
            self.database.insert_token(self.public_key, token)

            # If the metadata belongs to this token and chain, insert it.
            if metadata.verify(self.public_key) and metadata.token_pointer == token.get_hash():
                self.database.insert_metadata(self.public_key, metadata)

                # Insert all valid attestations to this metadata.
                valid_attestations = set()
                for authority_key, attestation in attestations:
                    if self.add_attestation(authority_key, attestation):
                        valid_attestations.add(attestation)

                out = Credential(metadata, valid_attestations)
                self.credentials.append(out)
                return out
        return None

    def add_attestation(self, public_key: PublicKey, attestation: Attestation) -> bool:
        """
        Add an attestation to this pseudonym.

        :returns: True on success, otherwise False
        """
        if attestation.verify(public_key):
            self.database.insert_attestation(self.public_key, public_key, attestation)
            return True
        return False

    def add_metadata(self, metadata: Metadata) -> bool:
        """
        Add a metadata entry to this pseudonym.

        :returns: True on success, otherwise False
        """
        if metadata.verify(self.public_key):
            self.database.insert_metadata(self.public_key, metadata)
            return True
        return False

    def create_attestation(self, metadata: Metadata, private_key: PrivateKey) -> Attestation:
        """
        Create an attestation for a a metadata entry of this pseudonym.

        Attesting to your own attributes is allowed (but probably not of any added value).
        """
        return Attestation.create(metadata, private_key)

    def create_credential(self,
                          attestation_hash: bytes,
                          metadata_json: dict,
                          after: typing.Optional[Metadata] = None) -> Credential:
        """
        Create a credential and add it to this pseudonym.
        """
        preceding = None if after is None else self.tree.elements.get(after.token_pointer, None)
        token = self.tree.add_by_hash(attestation_hash, preceding)
        metadata = Metadata(token.get_hash(), json.dumps(metadata_json).encode(), self.tree.private_key)
        return self.add_credential(token, metadata, set())  # type:ignore

    def get_credential(self, metadata: Metadata) -> Credential:
        """
        Return the credential belonging to given metadata.
        """
        return self.database.get_credential_over(metadata)

    def get_credentials(self) -> typing.List[Credential]:
        """
        Get all credentials belonging to this pseudonym.
        """
        return self.database.get_credentials_for(self.public_key)

    def disclose_credentials(self,
                             credentials: typing.Iterable[Credential],
                             attestation_selector: typing.Set[bytes]) -> typing.Tuple[bytes, bytes, bytes, bytes]:
        """
        Create a public disclosure for the given credentials and only include attestations from the given serialized
        public keys.

        This method automatically minimizes the amount of shared tokens and metadata.

        Another instance would receive and load these in using the IdentityManager's `substantiate()` method.

        :returns: the serialized metadata, tokens, attestations and authorities.
        """
        return self.create_disclosure({credential.metadata for credential in credentials}, attestation_selector)

    def create_disclosure(self,
                          metadata: typing.Set[Metadata],
                          attestation_selector: typing.Set[bytes]) -> typing.Tuple[bytes, bytes, bytes, bytes]:
        """
        Create a public disclosure for the given set of metadata and only include attestations from the given serialized
        public keys.

        This method automatically minimizes the amount of shared tokens and metadata.

        Another instance would receive and load these in using the IdentityManager's `substantiate()` method.

        :returns: the serialized metadata, tokens, attestations and authorities.
        """
        s_metadata = b''
        for md in metadata:
            serialized = md.get_plaintext_signed()
            s_metadata += struct.pack('>I', len(serialized)) + serialized
        attestations = b''
        authorities = b''
        for m in metadata:
            available_attestations = self.database.get_attestations_over(m)
            for attestation in available_attestations:
                if attestation.get_hash() in attestation_selector:
                    attestations += attestation.get_plaintext_signed()
                    authority = self.database.get_authority(attestation)
                    authority_len = len(authority)
                    authorities += struct.pack(">H", authority_len) + authority
        required_token_hashes = {m.token_pointer for m in metadata}
        tokens = set()
        # Keep adding tokens to this disclosure until all metadata is satisfied.
        for required_token_hash in required_token_hashes:
            root_token = self.tree.elements[required_token_hash]  # Throws a KeyError if the hash is unknown.
            # Avoid infinite loops and broken disclosures.
            if not self.tree.verify(root_token):
                raise RuntimeError("Attempted to create disclosure for undisclosable Token!")
            tokens.add(root_token)
            current_token = root_token
            while current_token.previous_token_hash != self.tree.genesis_hash:
                current_token = self.tree.elements[current_token.previous_token_hash]
                tokens.add(current_token)
        return s_metadata, b''.join(token.get_plaintext_signed() for token in tokens), attestations, authorities


class IdentityManager(object):
    """
    Manager of our own pseudonyms and those of others.
    """

    def __init__(self, database_path=":memory:"):
        super(IdentityManager, self).__init__()

        self.database = IdentityDatabase(database_path)
        self.database.open()
        self.pseudonyms = {}

        self.crypto = ECCrypto()

    def get_pseudonym(self, key: typing.Union[PublicKey, PrivateKey]) -> PseudonymManager:
        """
        Get the pseudonym belonging to a given public or private key.
        """
        public_key_material = key.pub().key_to_bin()
        if public_key_material not in self.pseudonyms:
            # Gotcha: PrivateKey is a subclass of PublicKey
            if isinstance(key, PrivateKey):
                self.pseudonyms[public_key_material] = PseudonymManager(self.database, private_key=key)
            else:
                self.pseudonyms[public_key_material] = PseudonymManager(self.database, public_key=key)
        return self.pseudonyms[public_key_material]

    def substantiate(self,
                     public_key: PublicKey,
                     serialized_metadata: bytes,
                     serialized_tokens: bytes,
                     serialized_attestations: bytes,
                     serialized_authorities: bytes) -> typing.Tuple[bool, PseudonymManager]:
        """
        Load the serialized form of a pseudonym for a given public key.

        :returns: whether this pseudonym is valid, the newly loaded pseudonym itself
        """
        # Load the tree structure
        pseudonym = self.get_pseudonym(public_key)
        correct = pseudonym.tree.unserialize_public(serialized_tokens)

        # Load the metadata
        metadata_offset = 0
        while metadata_offset < len(serialized_metadata):
            metadata_len, = struct.unpack_from('>I', serialized_metadata, metadata_offset)
            md = Metadata.unserialize(serialized_metadata[metadata_offset + 4:
                                                          metadata_offset + 4 + metadata_len], public_key)
            pseudonym.add_metadata(md)
            metadata_offset += 4 + metadata_len

        # Load the attestations and respective authorities
        attestation_offset = 0
        authority_offset = 0
        while authority_offset < len(serialized_authorities):
            authority_len, = struct.unpack_from('>H', serialized_authorities, authority_offset)
            authority = self.crypto.key_from_public_bin(serialized_authorities[authority_offset + 2:
                                                                               authority_offset + 2 + authority_len])
            authority_offset += 2 + authority_len
            correct &= pseudonym.add_attestation(authority,
                                                 Attestation.unserialize(serialized_attestations, authority,
                                                                         attestation_offset))
            attestation_offset += 32 + authority.get_signature_length()

        return correct, pseudonym



from asyncio import sleep

from ....base import TestBase
from .....messaging.interfaces.endpoint import EndpointListener
from .....messaging.interfaces.udp.endpoint import UDPEndpoint


class DummyEndpointListener(EndpointListener):
    """
    This class simply listens on an endpoint and stores incoming packets in a list.
    """
    def __init__(self, endpoint):
        super(DummyEndpointListener, self).__init__(endpoint)
        self.incoming = []

    def on_packet(self, packet):
        self.incoming.append(packet)


class TestUDPEndpoint(TestBase):
    """
    This class contains various tests for the UDP endpoint.
    """
    async def setUp(self):
        super(TestUDPEndpoint, self).setUp()
        self.endpoint1 = UDPEndpoint()
        await self.endpoint1.open()
        self.endpoint2 = UDPEndpoint()
        await self.endpoint2.open()

        self.ep2_address = ("127.0.0.1", self.endpoint2.get_address()[1])

        self.endpoint2_listener = DummyEndpointListener(self.endpoint2)
        self.endpoint2.add_listener(self.endpoint2_listener)

    async def tearDown(self):
        # If an endpoint was used, close it
        if self.endpoint1.is_open():
            self.endpoint1.close()
        if self.endpoint2.is_open():
            self.endpoint2.close()
        await super(TestUDPEndpoint, self).tearDown()

    async def test_send_message(self):
        """
        Test sending a basic message through the UDP endpoint.
        """
        # Send the package
        datum = b'a' * 10
        self.endpoint1.send(self.ep2_address, b'a' * 10)
        await sleep(.05)

        self.assertTrue(self.endpoint2_listener.incoming)
        self.assertEqual(self.endpoint2_listener.incoming[0][1], datum, "The received data was not the same as the"
                                                                        "sent data.")

    async def test_send_many_messages(self):
        """
        Test sending multiple messages through the UDP endpoint.
        """
        # range must be in [1, 51), since Asyncio transports discard empty datagrams
        for ind in range(1, 51):
            self.endpoint1.send(self.ep2_address, b'a' * ind)
        while len(self.endpoint2_listener.incoming) < 50:
            await sleep(.02)
        self.assertEqual(len(self.endpoint2_listener.incoming), 50)

    async def test_send_too_big_message(self):
        """
        Test sending a too big message through the UDP endpoint.
        """
        self.endpoint1.send(self.ep2_address, b'a' * (70000))
        await sleep(.05)
        self.assertFalse(self.endpoint2_listener.incoming)

    def test_send_invalid_destination(self):
        """
        Test sending a message with an invalid destination through the UDP endpoint.
        """
        self.endpoint1.send(("0.0.0.0", 0), b'a' * 10)



from random import sample
from time import time

from .discovery import DiscoveryStrategy


class RandomChurn(DiscoveryStrategy):
    """
    Select random peers, ping them if inactive, remove them if unresponsive.
    """

    def __init__(self, overlay, sample_size=8, ping_interval=10.0, inactive_time=27.5, drop_time=57.5):
        """
        Random peer removal strategy.

        :param overlay: the overlay to sample peers from
        :param sample_size: the amount of peers to check at once
        :param ping_interval: time between pings in the range of inactive_time to drop_time
        :param inactive_time: time before pings are sent to check liveness
        :param drop_time: time after which a peer is dropped
        """
        super(RandomChurn, self).__init__(overlay)
        self._pinged = {}
        self.sample_size = sample_size
        self.ping_interval = ping_interval
        self.inactive_time = inactive_time
        self.drop_time = drop_time

    def should_drop(self, peer):
        """
        Have we passed the time before we consider this peer to be unreachable.
        """
        if peer.last_response == 0:
            return False
        return time() > (peer.last_response + self.drop_time)

    def is_inactive(self, peer):
        """
        Have we passed the time before we consider this peer to be inactive.
        """
        if peer.last_response == 0:
            return False
        return time() > (peer.last_response + self.inactive_time)

    def take_step(self):
        """
        Select a new (set of) peer(s) to investigate liveness for.
        """
        with self.walk_lock:
            # Find an inactive or droppable peer
            sample_size = min(len(self.overlay.network.verified_peers), self.sample_size)
            if sample_size:
                window = sample(self.overlay.network.verified_peers, sample_size)

                for peer in window:
                    if self.should_drop(peer) and peer.address in self._pinged:
                        self.overlay.network.remove_peer(peer)
                        self._pinged.pop(peer.address)
                    elif self.is_inactive(peer) or len(peer.pings) < peer.pings.maxlen:
                        if ((peer.address in self._pinged)
                                and (time() > (self._pinged[peer.address] + self.ping_interval))):
                            self._pinged.pop(peer.address)
                        if peer.address not in self._pinged:
                            self._pinged[peer.address] = time()
                            self.overlay.send_ping(peer)



"""
The hidden tunnel community.

Author(s): Egbert Bouman
"""
import binascii
import os
import random
import socket
import struct
from asyncio import CancelledError, gather, iscoroutine

from .caches import *
from .community import TunnelCommunity, unpack_cell
from .payload import *
from .tunnel import (CIRCUIT_ID_PORT, CIRCUIT_TYPE_IP_SEEDER, CIRCUIT_TYPE_RP_DOWNLOADER, CIRCUIT_TYPE_RP_SEEDER,
                     DESTROY_REASON_LEAVE_SWARM, DESTROY_REASON_UNNEEDED, EXIT_NODE, EXIT_NODE_SALT, Hop,
                     IntroductionPoint, PEER_SOURCE_DHT, PEER_SOURCE_PEX, RelayRoute, RendezvousPoint, Swarm,
                     TunnelExitSocket)
from ...bootstrapping.dispersy.bootstrapper import DispersyBootstrapper
from ...configuration import DISPERSY_BOOTSTRAPPER
from ...keyvault.public.libnaclkey import LibNaCLPK
from ...messaging.anonymization.pex import PexCommunity
from ...peer import Peer
from ...peerdiscovery.churn import RandomChurn
from ...peerdiscovery.discovery import RandomWalk
from ...peerdiscovery.network import Network
from ...taskmanager import task
from ...util import fail


class HiddenTunnelCommunity(TunnelCommunity):

    def __init__(self, *args, **kwargs):
        self.ipv8 = kwargs.pop('ipv8', None)
        self.e2e_callbacks = kwargs.pop('e2e_callbacks', {})

        self.swarms = {}
        self.pex = {}

        super(HiddenTunnelCommunity, self).__init__(*args, **kwargs)

        self.intro_point_for = {}  # {seeder_pk: (TunnelExitSocket, info_hash)}
        self.rendezvous_point_for = {}  # {cookie: TunnelExitSocket}

        # Messages that can arrive from the socket
        self.add_message_handler(CreateE2EPayload, self.on_create_e2e)
        self.add_message_handler(PeersRequestPayload, self.on_peers_request)
        self.add_message_handler(PeersResponsePayload, self.on_peers_response)

        # Messages that can arrive from a circuit (i.e., they are wrapped in a cell)
        self.add_cell_handler(EstablishIntroPayload, self.on_establish_intro)
        self.add_cell_handler(IntroEstablishedPayload, self.on_intro_established)
        self.add_cell_handler(EstablishRendezvousPayload, self.on_establish_rendezvous)
        self.add_cell_handler(RendezvousEstablishedPayload, self.on_rendezvous_established)
        self.add_cell_handler(CreateE2EPayload, self.on_create_e2e)
        self.add_cell_handler(CreatedE2EPayload, self.on_created_e2e)
        self.add_cell_handler(LinkE2EPayload, self.on_link_e2e)
        self.add_cell_handler(LinkedE2EPayload, self.on_linked_e2e)
        self.add_cell_handler(PeersRequestPayload, self.on_peers_request)
        self.add_cell_handler(PeersResponsePayload, self.on_peers_response)

        self.register_task("do_peer_discovery", self.do_peer_discovery, interval=10)

    def join_swarm(self, info_hash, hops, callback=None, seeding=True):
        """
        Join a hidden swarm. This should be called by both the downloader and the seeder. Calling this method while
        already part of the swarm will cause the community to drop all pre-existing connections.
        Note that the seeder should also create introduction points by calling create_introduction_point().

        :param info_hash: the swarm identifier
        :type info_hash: bytes
        :param hops: the amount of hops for our introduction/rendezvous circuits
        :type hops: int
        :param callback: the callback function to call when we have established an e2e circuit
        :param seeding: whether or not the swarm should be joined as seeder
        :type seeding: bool
        """
        if info_hash in self.swarms:
            self.logger.warning('Already part of hidden swarm %s, leaving existing', binascii.hexlify(info_hash))
            self.leave_swarm(info_hash)

        self.swarms[info_hash] = Swarm(info_hash, hops, self.send_peers_request,
                                       self.crypto.generate_key("curve25519") if seeding else None)
        self.e2e_callbacks[info_hash] = callback

    def leave_swarm(self, info_hash):
        """
        Leave a hidden swarm. Can be called by both the downloader and the seeder.

        :param info_hash: the swarm identifier
        :type info_hash: bytes
        """
        # Remove all introduction points and e2e circuits for this swarm
        for circuit in self.circuits.values():
            if circuit.info_hash == info_hash and circuit.ctype in [CIRCUIT_TYPE_IP_SEEDER,
                                                                    CIRCUIT_TYPE_RP_SEEDER,
                                                                    CIRCUIT_TYPE_RP_DOWNLOADER]:
                self.remove_circuit(circuit.circuit_id, 'leaving hidden swarm', destroy=DESTROY_REASON_LEAVE_SWARM)
        # Remove swarm and callback
        swarm = self.swarms.pop(info_hash, None)
        self.e2e_callbacks.pop(info_hash, None)
        # If there are no other swarms with the same hop count, remove the data circuits
        if swarm and not [s for s in self.swarms.values() if s != swarm and s.hops == swarm.hops]:
            for circuit in self.find_circuits(hops=swarm.hops, state=None):
                self.remove_circuit(circuit.circuit_id, 'not needed', destroy=DESTROY_REASON_UNNEEDED)

    async def estimate_swarm_size(self, info_hash, hops=1, max_requests=10):
        """
        Estimate the number of unique seeders that are part of a hidden swarm.

        :param info_hash: the swarm identifier
        :type info_hash: str
        :param hops: the amount of hops to use for contacting introduction points
        :type hops: int
        :param max_requests: the number of introduction points we should send a get-peers message to
        :type max_requests: int
        :return: number of unique seeders
        """
        swarm = Swarm(info_hash, hops, self.send_peers_request)

        # None represents a DHT request
        all_ = [None]
        tried = set()

        while len(tried) < max_requests:
            not_tried = set(all_) - tried
            if not not_tried:
                break

            # Issue as many requests as possible in parallel
            ips = random.sample(not_tried, min(len(not_tried), max_requests - len(tried)))
            responses = await gather(*[swarm.lookup(ip) for ip in ips], return_exceptions=True)

            # Collect responses
            all_ += sum([result for result in responses if not isinstance(result, (CancelledError, Exception))], [])
            tried |= set(ips)

        return len({ip.seeder_pk for ip in all_ if ip and ip.source == PEER_SOURCE_PEX})

    def select_circuit_for_infohash(self, info_hash):
        swarm = self.swarms.get(info_hash)
        if not swarm or not swarm.hops:
            self.logger.info('Can\'t get hop count, download cancelled?')
            return

        return self.select_circuit(None, swarm.hops)

    def create_circuit_for_infohash(self, info_hash, ctype, *args, **kwargs):
        swarm = self.swarms.get(info_hash)
        if not swarm or not swarm.hops:
            self.logger.info('Can\'t get hop count, download cancelled?')
            return

        # Introduction point circuits need an additional hop, or we will talk directly to the the introduction point.
        # Also, rendezvous circuits need an additional hop, since the seeder chooses the rendezvous_point.
        hops = swarm.hops
        if ctype in [CIRCUIT_TYPE_IP_SEEDER, CIRCUIT_TYPE_RP_DOWNLOADER]:
            hops += 1

        return self.create_circuit(hops, ctype, *args, info_hash=info_hash, **kwargs)

    def ip_to_circuit_id(self, ip_str):
        return struct.unpack("!I", socket.inet_aton(ip_str))[0]

    def circuit_id_to_ip(self, circuit_id):
        return socket.inet_ntoa(struct.pack("!I", circuit_id))

    async def do_peer_discovery(self):
        now = time.time()
        for info_hash, swarm in list(self.swarms.items()):
            if not swarm.seeding and swarm.last_lookup + self.settings.swarm_lookup_interval <= now \
               and swarm.get_num_connections() < self.settings.swarm_connection_limit:
                try:
                    ips = await swarm.lookup()
                except (IndexError, RuntimeError):
                    self.logger.info('Failed to do peer discovery for swarm %s', binascii.hexlify(info_hash))
                    continue

                if ips is None:
                    self.logger.info('Skipping peer discovery for swarm %s', binascii.hexlify(info_hash))
                    continue
                self.logger.info('Found %d/%d peer(s) for swarm %s',
                                 len([ip for ip in ips if ip.source == PEER_SOURCE_DHT]),
                                 len([ip for ip in ips if ip.source == PEER_SOURCE_PEX]),
                                 binascii.hexlify(info_hash))
                for ip in set(ips):
                    swarm.add_intro_point(ip)
                for ip in swarm.intro_points:
                    if not swarm.has_connection(ip.seeder_pk):
                        self.create_e2e(info_hash, ip)

    def do_circuits(self):
        super(HiddenTunnelCommunity, self).do_circuits()

        # Make sure we have at least 1 data circuit for every hop count. This circuit will be used for
        # communication with introduction points.
        for hop_count in {swarm.hops for swarm in self.swarms.values()}:
            if not self.find_circuits(state=None, hops=hop_count):
                self.create_circuit(hop_count)

    def do_ping(self, exclude=None):
        # Ping all circuits, except pending e2e circuits
        exclude = [] if exclude is None else exclude
        exclude += [c.circuit_id for c in self.circuits.values()
                    if (c.ctype == CIRCUIT_TYPE_RP_SEEDER) or (c.ctype == CIRCUIT_TYPE_RP_DOWNLOADER and not c.e2e)]
        super(HiddenTunnelCommunity, self).do_ping(exclude=exclude)

    def remove_circuit(self, circuit_id, additional_info='', remove_now=False, destroy=False):
        circuit = self.circuits.get(circuit_id, None)
        if circuit and circuit.ctype == CIRCUIT_TYPE_RP_DOWNLOADER:
            swarm = self.swarms.get(circuit.info_hash)
            if swarm:
                swarm.remove_connection(circuit)
        return super(HiddenTunnelCommunity, self).remove_circuit(circuit_id, additional_info, remove_now, destroy)

    def remove_exit_socket(self, circuit_id, additional_info='', remove_now=False, destroy=False):
        for seeder_pk, (intro_circuit, info_hash) in list(self.intro_point_for.items()):
            if intro_circuit.circuit_id == circuit_id:
                self.intro_point_for.pop(seeder_pk)

                # Stop announcing in PEX community
                pex = self.pex.get(info_hash)
                if pex:
                    pex.stop_announce(seeder_pk)

                    # Unload PEX community
                    if pex.done:
                        self.pex.pop(info_hash, None)
                        self.ipv8.overlays.remove(pex)
                        self.ipv8.strategies = [t for t in self.ipv8.strategies if t[0].overlay != pex]
                        self.register_anonymous_task('unload_pex', pex.unload)

        for cookie, rendezvous_circuit in list(self.rendezvous_point_for.items()):
            if rendezvous_circuit.circuit_id == circuit_id:
                self.rendezvous_point_for.pop(cookie)

        return super(HiddenTunnelCommunity, self).remove_exit_socket(circuit_id, additional_info, remove_now, destroy)

    def get_max_time(self, circuit_id):
        if circuit_id in self.circuits and self.circuits[circuit_id].ctype == CIRCUIT_TYPE_IP_SEEDER:
            return self.settings.max_time_ip
        if circuit_id in self.exit_sockets and circuit_id in [c.circuit_id for c, _ in self.intro_point_for.values()]:
            return self.settings.max_time_ip
        return super(HiddenTunnelCommunity, self).get_max_time(circuit_id)

    def tunnel_data(self, circuit, destination, payload):
        packet = self._ez_pack(self._prefix, payload.msg_id, [payload], False)
        pre = ('0.0.0.0', 0)
        post = ('0.0.0.0', 0)
        if isinstance(circuit, TunnelExitSocket):
            post = destination
        else:
            pre = destination
        self.send_data(circuit.peer, circuit.circuit_id, pre, post, packet)

    def select_circuit(self, destination, hops):
        # Make sure that we select the right circuit when dealing with an e2e connection
        if destination and destination[1] == CIRCUIT_ID_PORT:
            circuit_id = self.ip_to_circuit_id(destination[0])
            circuit = self.circuits.get(circuit_id, None)

            if circuit and circuit.state == CIRCUIT_STATE_READY and \
               circuit.ctype == CIRCUIT_TYPE_RP_DOWNLOADER:
                return circuit

        circuits = self.find_circuits(hops=hops)
        return random.choice(circuits) if circuits else None

    def send_peers_request(self, info_hash, target, hops):
        circuit = self.select_circuit(None, hops)
        if not circuit:
            self.logger.info("No circuit for peers-request")
            return fail(RuntimeError("No circuit for peers-request"))

        # Send a peers-request message over this circuit
        cache = PeersRequestCache(self, circuit, info_hash, target)
        self.request_cache.add(cache)
        payload = PeersRequestPayload(circuit.circuit_id, cache.number, info_hash)

        # Ask an introduction point if available (in which case we'll use PEX), otherwise let
        # the exit node do a DHT request.
        if target and target.peer.public_key != circuit.hops[-1].public_key:
            self.tunnel_data(circuit, target.peer.address, payload)
            self.logger.info("Sending peers request (intro point %s)", target.peer)
        else:
            self.send_cell(circuit.peer, payload)
            self.logger.info("Sending peers request as cell")
        return cache.future

    @unpack_cell(PeersRequestPayload)
    async def on_peers_request(self, source_address, payload, circuit_id=None):
        info_hash = payload.info_hash
        self.logger.info("Doing hidden seeders lookup for info_hash %s", binascii.hexlify(info_hash))
        if info_hash in self.pex:
            # Get peers from PEX community
            intro_points = self.pex[info_hash].get_intro_points()
            self.send_peers_response(source_address, payload, intro_points, circuit_id)
        elif circuit_id in self.exit_sockets:
            # Get peers from DHT community
            results = await self.dht_lookup(info_hash)
            if results:
                _, intro_points = results
                self.send_peers_response(source_address, payload, intro_points, circuit_id)
        elif circuit_id is not None:
            self.logger.warning("Received a peers-request over circuit %d, but unable to do a DHT lookup", circuit_id)
        else:
            self.logger.warning("Received a peers-request over the socket, but unable to do a PEX lookup")

    def send_peers_response(self, target_addr, request, intro_points, circuit_id):
        peers = [IntroductionInfo(ip.peer.address, ip.peer.public_key.key_to_bin(), ip.seeder_pk, ip.source)
                 for ip in random.sample(intro_points, min(len(intro_points), 7))]
        payload = PeersResponsePayload(request.circuit_id, request.identifier, request.info_hash, peers)

        if circuit_id is not None:
            # Send back to origin
            self.send_cell(target_addr, payload)
        else:
            # Send back to exit node
            packet = self._ez_pack(self._prefix, payload.msg_id, [payload], False)
            self.send_packet(target_addr, packet)

    @unpack_cell(PeersResponsePayload)
    def on_peers_response(self, source_address, payload, circuit_id):
        if not self.request_cache.has("peers-request", payload.identifier):
            self.logger.warning('Got a peers-response with an unknown identifier')
            return
        cache = self.request_cache.pop("peers-request", payload.identifier)

        self.logger.info("Received peers-response containing %d peers", len(payload.peers))
        ips = [IntroductionPoint(Peer(peer.key, address=peer.address), peer.seeder_pk, peer.source)
               for peer in payload.peers if peer.address != ('0.0.0.0', 0)]
        cache.future.set_result(ips)

    def create_e2e(self, info_hash, intro_point):
        circuit = self.select_circuit_for_infohash(info_hash)
        if not circuit:
            self.logger.error("No circuit for contacting the introduction point")
            return

        hop = Hop(Peer(LibNaCLPK(intro_point.seeder_pk[10:])))
        hop.dh_secret, hop.dh_first_part = self.crypto.generate_diffie_secret()
        self.logger.info('Creating e2e circuit for introduction point %s', intro_point.peer)
        cache = E2ERequestCache(self, info_hash, hop, intro_point)
        self.request_cache.add(cache)
        self.tunnel_data(circuit, intro_point.peer.address,
                         CreateE2EPayload(cache.number, info_hash, hop.node_public_key, hop.dh_first_part))

    @unpack_cell(CreateE2EPayload)
    async def on_create_e2e(self, source_address, payload, circuit_id=None):
        # If we have received this message over a socket, we need to forward it
        if circuit_id is None:
            if payload.node_public_key in self.intro_point_for:
                self.logger.info('On create-e2e: forwarding message because received over socket')
                relay_circuit, _ = self.intro_point_for[payload.node_public_key]
                self.tunnel_data(relay_circuit, source_address, payload)
            else:
                self.logger.info('On create-e2e: dropping message for unknown seeder key %s',
                                 binascii.hexlify(payload.node_public_key))
        else:
            self.logger.info('On create-e2e: creating rendezvous point')
            swarm = self.swarms.get(payload.info_hash)
            if swarm and swarm.seeding:
                rp = await self.create_rendezvous_point(payload.info_hash)
                if rp and await rp.ready:
                    self.create_created_e2e(rp, source_address, payload, circuit_id)

    def create_created_e2e(self, rp, source_address, payload, circuit_id):
        key = self.swarms[payload.info_hash].seeder_sk
        shared_secret, Y, AUTH = self.crypto.generate_diffie_shared_secret(payload.key, key)
        rp.circuit.hs_session_keys = self.crypto.generate_session_keys(shared_secret)

        rp_info = RendezvousInfo(rp.address, rp.circuit.hops[-1].public_key.key_to_bin(), rp.cookie)
        rp_info_bin = self.serializer.pack('payload', rp_info)
        rp_info_enc = self.crypto.encrypt_str(rp_info_bin,
                                              *self.crypto.get_session_keys(rp.circuit.hs_session_keys, EXIT_NODE))

        circuit = self.circuits[circuit_id]
        self.tunnel_data(circuit, source_address, CreatedE2EPayload(payload.identifier, Y, AUTH, rp_info_enc))

    @unpack_cell(CreatedE2EPayload)
    async def on_created_e2e(self, source_address, payload, circuit_id):
        if not self.request_cache.has("e2e-request", payload.identifier):
            self.logger.warning("Invalid created-e2e identifier")
            return

        cache = self.request_cache.pop("e2e-request", payload.identifier)
        shared_secret = self.crypto.verify_and_generate_shared_secret(cache.hop.dh_secret,
                                                                      payload.key,
                                                                      payload.auth,
                                                                      cache.hop.public_key.key.pk)
        session_keys = self.crypto.generate_session_keys(shared_secret)

        rp_info_enc = payload.rp_info_enc
        rp_info_bin = self.crypto.decrypt_str(rp_info_enc, session_keys[EXIT_NODE], session_keys[EXIT_NODE_SALT])
        rp_info, _ = self.serializer.unpack(RendezvousInfo, rp_info_bin)

        required_exit = Peer(rp_info.key, rp_info.address)
        circuit = self.create_circuit_for_infohash(cache.info_hash, CIRCUIT_TYPE_RP_DOWNLOADER,
                                                   required_exit=required_exit)
        if circuit:
            self.swarms[cache.info_hash].add_connection(circuit, cache.intro_point)
            if await circuit.ready:
                cache = LinkRequestCache(self, circuit, cache.info_hash, session_keys)
                self.request_cache.add(cache)
                self.send_cell(circuit.peer, LinkE2EPayload(circuit.circuit_id, cache.number, rp_info.cookie))

    @unpack_cell(LinkE2EPayload)
    def on_link_e2e(self, source_address, payload, circuit_id):
        if payload.cookie not in self.rendezvous_point_for:
            self.logger.warning("Not a rendezvous point for this cookie")
            return

        if self.exit_sockets[circuit_id].enabled:
            self.logger.warning("Exit socket for circuit is enabled, cannot link")
            return

        relay_circuit = self.rendezvous_point_for[payload.cookie]
        if self.exit_sockets[relay_circuit.circuit_id].enabled:
            self.logger.warning("Exit socket for relay_circuit is enabled, cannot link")
            return

        circuit = self.exit_sockets[circuit_id]

        self.remove_exit_socket(circuit.circuit_id, 'linking circuit')
        self.remove_exit_socket(relay_circuit.circuit_id, 'linking circuit')

        self.relay_from_to[circuit.circuit_id] = RelayRoute(relay_circuit.circuit_id, relay_circuit.peer, True)
        self.relay_from_to[relay_circuit.circuit_id] = RelayRoute(circuit.circuit_id, circuit.peer, True)

        self.send_cell(source_address, LinkedE2EPayload(circuit.circuit_id, payload.identifier))

    @unpack_cell(LinkedE2EPayload)
    def on_linked_e2e(self, source_address, payload, circuit_id):
        if not self.request_cache.has("link-request", payload.identifier):
            self.logger.warning("Invalid linked-e2e identifier")
            return

        cache = self.request_cache.pop("link-request", payload.identifier)
        circuit = cache.circuit
        circuit.e2e = True
        circuit.hs_session_keys = cache.hs_session_keys
        callback = self.e2e_callbacks.get(cache.info_hash, None)
        if callback:
            result = callback((self.circuit_id_to_ip(circuit.circuit_id), CIRCUIT_ID_PORT))
            if iscoroutine(result):
                self.register_anonymous_task('e2e_callback', result)
        else:
            self.logger.error('On linked e2e: could not find download for %s!', cache.info_hash)

    async def create_introduction_point(self, info_hash, required_ip=None):
        self.logger.info("Creating introduction point")

        if info_hash not in self.swarms:
            self.logger.warning('Cannot create introduction point for unknown swarm')
            return
        elif not self.swarms[info_hash].seeding:
            self.logger.warning('Cannot create introduction point for swarm that is not seeding')
            return

        circuit = self.create_circuit_for_infohash(info_hash, CIRCUIT_TYPE_IP_SEEDER, required_exit=required_ip)

        if circuit and await circuit.ready:
            # We got a circuit, now let's create an introduction point
            seed_pk = self.swarms[info_hash].seeder_sk.pub().key_to_bin()
            circuit_id = circuit.circuit_id
            cache = IPRequestCache(self, circuit)
            self.request_cache.add(cache)
            self.send_cell(circuit.peer, EstablishIntroPayload(circuit_id, cache.number, info_hash, seed_pk))
            self.logger.info("Established introduction tunnel %s", circuit_id)

    @unpack_cell(EstablishIntroPayload)
    def on_establish_intro(self, source_address, payload, circuit_id):
        if payload.public_key in self.intro_point_for:
            self.logger.warning('Already have an introduction point for %s', binascii.hexlify(payload.public_key))
            return

        self.logger.info('Established introduction point for %s', binascii.hexlify(payload.public_key))

        circuit = self.exit_sockets[circuit_id]
        self.intro_point_for[payload.public_key] = circuit, payload.info_hash

        if not self.ipv8:
            self.logger.error('No IPv8 service object available, cannot start PEXCommunity')
        elif payload.info_hash not in self.pex:
            community = PexCommunity(self.my_peer, self.endpoint, Network(), info_hash=payload.info_hash)
            community.bootstrappers = [DispersyBootstrapper(**DISPERSY_BOOTSTRAPPER['init'])]
            # Since IPv8 takes a step every .5s until we have 10 peers, the PexCommunity will generate
            # a lot of traffic in case there are <10 peers in existence. Therefore, we slow the walk down to a 5s/step.
            self.ipv8.add_strategy(community, RandomWalk(community, target_interval=5), 10)
            self.ipv8.add_strategy(community, RandomChurn(community), -1)
            self.pex[payload.info_hash] = community

        # PEX announce
        if payload.info_hash in self.pex:
            self.pex[payload.info_hash].start_announce(payload.public_key)

        # DHT announce
        self.dht_announce(payload.info_hash, IntroductionPoint(Peer(self.my_peer.key, self.my_estimated_wan),
                                                               payload.public_key))

        self.send_cell(source_address, IntroEstablishedPayload(circuit.circuit_id, payload.identifier))

    @unpack_cell(IntroEstablishedPayload)
    def on_intro_established(self, source_address, payload, circuit_id):
        if not self.request_cache.has("establish-intro", payload.identifier):
            self.logger.warning("Invalid intro-established request identifier")
            return

        self.request_cache.pop("establish-intro", payload.identifier)
        self.logger.info("Got intro-established from %s", source_address)

    async def create_rendezvous_point(self, info_hash):
        # Create a new circuit to be used for transferring data
        circuit = self.create_circuit_for_infohash(info_hash, CIRCUIT_TYPE_RP_SEEDER)

        if circuit and await circuit.ready:
            # We got a circuit, now let's create a rendezvous point
            rp = RendezvousPoint(circuit, os.urandom(20))
            cache = RPRequestCache(self, rp)
            self.request_cache.add(cache)
            self.send_cell(circuit.peer, EstablishRendezvousPayload(circuit.circuit_id, cache.number, rp.cookie))
            return rp

    @unpack_cell(EstablishRendezvousPayload)
    def on_establish_rendezvous(self, source_address, payload, circuit_id):
        circuit = self.exit_sockets[circuit_id]
        self.rendezvous_point_for[payload.cookie] = circuit

        self.send_cell(source_address,
                       RendezvousEstablishedPayload(circuit.circuit_id, payload.identifier, self.my_estimated_wan))

    @unpack_cell(RendezvousEstablishedPayload)
    def on_rendezvous_established(self, source_address, payload, circuit_id):
        if not self.request_cache.has("establish-rendezvous", payload.identifier):
            self.logger.warning("Invalid rendezvous-established request identifier")
            return

        rp = self.request_cache.pop("establish-rendezvous", payload.identifier).rp
        rp.address = payload.rendezvous_point_addr
        rp.ready.set_result(rp)

    async def dht_lookup(self, info_hash):
        if self.dht_provider:
            return await self.dht_provider.lookup(info_hash)
        else:
            self.logger.error("Need a DHT provider to lookup on the DHT")

    @task
    async def dht_announce(self, info_hash, intro_point):
        if self.dht_provider:
            return await self.dht_provider.announce(info_hash, intro_point)
        else:
            self.logger.error("Need a DHT provider to announce to the DHT")



from base64 import b64encode

from .base import TestBase
from ..keyvault.crypto import default_eccrypto
from ..messaging.interfaces.udp.endpoint import UDPv4Address, UDPv6Address
from ..peer import Peer


class TestPeer(TestBase):

    test_key = default_eccrypto.generate_key(u"very-low")

    def setUp(self):
        super(TestPeer, self).setUp()
        self.peer = Peer(TestPeer.test_key, ("1.2.3.4", 5))

    def test_default_timestamp(self):
        """
        Check if the default Lamport timestamp of a Peer is 0.
        """
        self.assertEqual(self.peer.get_lamport_timestamp(), 0)

    def test_increment_timestamp(self):
        """
        Check if the Lamport timestamp of a Peer can be incremented.
        """
        self.peer.update_clock(1)

        self.assertEqual(self.peer.get_lamport_timestamp(), 1)

    def test_increase_timestamp(self):
        """
        Check if the Lamport timestamp of a Peer can be increased arbitrarily.
        """
        self.peer.update_clock(42)

        self.assertEqual(self.peer.get_lamport_timestamp(), 42)

    def test_decrease_timestamp(self):
        """
        Check if the Lamport timestamp of a Peer cannot be decreased.
        """
        self.peer.update_clock(-1)

        self.assertEqual(self.peer.get_lamport_timestamp(), 0)

    def test_peer_equality(self):
        """
        Check if peers with the same key and address are equal.
        """
        other = Peer(self.peer.key, self.peer.address)

        self.assertTrue(self.peer == other)
        self.assertFalse(self.peer != other)

    def test_peer_inequality_key(self):
        """
        Check if peers with a different key and same address are not equal.
        """
        other = Peer(default_eccrypto.generate_key(u"very-low"), self.peer.address)

        self.assertNotEqual(self.peer, other)

    def test_median_ping_none(self):
        """
        No ping measurements should lead to a None median ping.
        """
        self.assertIsNone(self.peer.get_median_ping())

    def test_avg_ping_none(self):
        """
        No ping measurements should lead to a None average ping.
        """
        self.assertIsNone(self.peer.get_average_ping())

    def test_median_ping_odd(self):
        """
        Median ping should return the median ping for odd length measurements.
        """
        self.peer.pings.append(2.0)
        self.peer.pings.append(3.0)
        self.peer.pings.append(4.0)
        self.assertEqual(3.0, self.peer.get_median_ping())

    def test_median_ping_even(self):
        """
        Median ping should return the median ping for even length measurements.
        """
        self.peer.pings.append(2.0)
        self.peer.pings.append(3.0)
        self.peer.pings.append(4.0)
        self.peer.pings.append(5.0)
        self.assertEqual(3.5, self.peer.get_median_ping())

    def test_avg_ping(self):
        """
        Average ping should return the average ping.
        """
        self.peer.pings.append(3.0)
        self.peer.pings.append(4.0)
        self.assertEqual(3.5, self.peer.get_average_ping())

    def test_peer_inequality_address(self):
        """
        Check if peers with the same key and a different address are equal.
        """
        other = Peer(self.peer.key)

        self.assertEqual(self.peer, other)

    def test_to_string(self):
        """
        Check if the __str__ method functions properly.
        """
        self.assertEqual(str(self.peer), "Peer<1.2.3.4:5, %s>" % b64encode(self.peer.mid).decode('utf-8'))

    def test_set_address_init(self):
        """
        Check if the address property properly sets from the init.
        """
        address = UDPv4Address("1.2.3.4", 5)
        peer = Peer(TestPeer.test_key, address)

        self.assertEqual(peer.address, address)

    def test_set_address_setter(self):
        """
        Check if the address property properly sets from the setter.
        """
        address = UDPv4Address("1.2.3.4", 5)
        peer = Peer(TestPeer.test_key)
        peer.address = address

        self.assertEqual(peer.address, address)

    def test_set_address_add(self):
        """
        Check if the address property properly sets from add_address.
        """
        address = UDPv4Address("1.2.3.4", 5)
        peer = Peer(TestPeer.test_key)
        peer.add_address(address)

        self.assertEqual(peer.address, address)

    def test_set_address_addv6(self):
        """
        Check if IPv6 addresses are properly returned.
        """
        address = UDPv6Address("1:2:3:4:5:6", 7)
        peer = Peer(TestPeer.test_key)
        peer.add_address(address)

        self.assertEqual(peer.address, address)

    def test_address_order1(self):
        """
        Check if IPv6 is preferred over IPv4 (append out-of-order).
        """
        address1 = UDPv4Address("1.2.3.4", 5)
        address2 = UDPv6Address("1:2:3:4:5:6", 7)
        peer = Peer(TestPeer.test_key)
        peer.add_address(address2)
        peer.add_address(address1)

        self.assertEqual(peer.address, address2)

    def test_address_order2(self):
        """
        Check if IPv6 is preferred over IPv4 (append in-order).
        """
        address1 = UDPv4Address("1.2.3.4", 5)
        address2 = UDPv6Address("1:2:3:4:5:6", 7)
        peer = Peer(TestPeer.test_key)
        peer.add_address(address1)
        peer.add_address(address2)

        self.assertEqual(peer.address, address2)

    def test_default_address(self):
        """
        Check if the default address is UDPv4Address("0.0.0.0", 0)
        """
        self.assertEqual(Peer(TestPeer.test_key).address, UDPv4Address("0.0.0.0", 0))

    def test_manual_update(self):
        """
        Check if manual updates to the addresses dictionary are caught.
        """
        address = UDPv4Address("1.2.3.4", 5)
        peer = Peer(TestPeer.test_key)
        peer.addresses.update({UDPv4Address: address})

        self.assertEqual(peer.address, address)

    def test_manual_updates(self):
        """
        Check if manual updates to the addresses dictionary are caught (double update, out-of-order).
        """
        address1 = UDPv6Address("1:2:3:4:5:6", 7)
        address2 = UDPv4Address("1.2.3.4", 5)
        peer = Peer(TestPeer.test_key)
        peer.addresses.update({UDPv4Address: address2})
        peer.addresses.update({UDPv6Address: address1})

        self.assertEqual(peer.address, address1)

    def test_manual_update_overwrite(self):
        """
        Check if manual updates to the addresses dictionary are caught (overwrite same class).
        """
        address = UDPv4Address("1.2.3.4", 5)
        peer = Peer(TestPeer.test_key, UDPv4Address("6.7.8.9", 10))
        peer.addresses.update({UDPv4Address: address})

        self.assertEqual(peer.address, address)



import json
from asyncio import Future
from base64 import b64decode, b64encode
from hashlib import sha1

from aiohttp import web

from aiohttp_apispec import docs

from .base_endpoint import BaseEndpoint, HTTP_BAD_REQUEST, HTTP_NOT_FOUND, Response
from ..attestation.identity.community import IdentityCommunity, create_community
from ..attestation.wallet.community import AttestationCommunity
from ..keyvault.crypto import default_eccrypto
from ..peer import Peer
from ..util import strip_sha1_padding, succeed


class AttestationEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handing all requests regarding attestation.
    """

    def __init__(self):
        super(AttestationEndpoint, self).__init__()
        self.attestation_overlay = self.identity_overlay = self.persistent_key = None
        self.attestation_requests = {}
        self.verify_requests = {}
        self.verification_output = {}
        self.attestation_metadata = {}

    def setup_routes(self):
        self.app.add_routes([web.get('', self.handle_get),
                             web.post('', self.handle_post)])

    def initialize(self, session):
        super(AttestationEndpoint, self).initialize(session)
        self.attestation_overlay = next((overlay for overlay in session.overlays
                                         if isinstance(overlay, AttestationCommunity)), None)
        self.identity_overlay = next((overlay for overlay in session.overlays
                                      if isinstance(overlay, IdentityCommunity)), None)
        if self.attestation_overlay and self.identity_overlay:
            self.attestation_overlay.set_attestation_request_callback(self.on_request_attestation)
            self.attestation_overlay.set_attestation_request_complete_callback(self.on_attestation_complete)
            self.attestation_overlay.set_verify_request_callback(self.on_verify_request)
            self.persistent_key = self.identity_overlay.my_peer

    def on_request_attestation(self, peer, attribute_name, metadata):
        """
        Return the measurement of an attribute for a certain peer.
        """
        future = Future()
        self.attestation_requests[(b64encode(peer.mid).decode(), attribute_name)] = \
            (future, b64encode(json.dumps(metadata).encode('utf-8')).decode())
        self.attestation_metadata[(peer, attribute_name)] = metadata
        return future

    def on_attestation_complete(self, for_peer, attribute_name, attribute_hash, id_format, from_peer=None):
        """
        Callback for when an attestation has been completed for another peer.
        We can now sign for it.
        """
        metadata = self.attestation_metadata.get((for_peer, attribute_name), None)
        if for_peer.mid == self.identity_overlay.my_peer.mid:
            if from_peer.mid == self.identity_overlay.my_peer.mid:
                self.identity_overlay.self_advertise(attribute_hash, attribute_name, id_format, metadata)
            else:
                self.identity_overlay.request_attestation_advertisement(from_peer, attribute_hash, attribute_name,
                                                                        id_format, metadata)
        else:
            self.identity_overlay.add_known_hash(attribute_hash, attribute_name, for_peer.public_key.key_to_bin(),
                                                 metadata)

    def on_verify_request(self, peer, attribute_hash):
        """
        Return the measurement of an attribute for a certain peer.
        """
        metadata = self.identity_overlay.get_attestation_by_hash(attribute_hash)
        if not metadata:
            return succeed(None)
        attribute_name = json.loads(metadata.serialized_json_dict)["name"]
        future = Future()
        self.verify_requests[(b64encode(peer.mid).decode(), attribute_name)] = future
        return future

    def on_verification_results(self, attribute_hash, values):
        """
        Callback for when verification has concluded.
        """
        references = self.verification_output[attribute_hash]
        out = []
        for i in range(len(references)):
            out.append((references[i][0] if isinstance(references[i], tuple) else references[i], values[i]))
        self.verification_output[attribute_hash] = out

    def get_peer_from_mid(self, mid_b64):
        """
        Find a peer by base64 encoded mid.
        """
        mid = b64decode(mid_b64)
        peers = self.session.network.verified_peers
        matches = [p for p in peers if p.mid == mid]
        return matches[0] if matches else None

    def _drop_identity_table_data(self, keys_to_keep):
        """
        Remove all metadata  from the identity community.

        :param keys_to_keep: list of keys to not remove for
        :type keys_to_keep: [str]
        :return: the list of attestation hashes which have been removed
        :rtype: [database_blob]
        """
        database = self.identity_overlay.identity_manager.database
        all_identities = database.get_known_identities()
        to_remove = []
        attestation_hashes = []
        for key in all_identities:
            if key not in keys_to_keep:
                to_remove.append(key)
                attestation_hashes.extend([t.content_hash for t in database.get_tokens_for(Peer(key).public_key)])

        with database:
            for public_key in to_remove:
                database.executescript("BEGIN TRANSACTION; "
                                       "DELETE FROM Tokens WHERE public_key = ?; "
                                       "DELETE FROM Metadata WHERE public_key = ?; "
                                       "DELETE FROM Attestations WHERE public_key = ?; "
                                       "COMMIT;",
                                       (public_key, public_key, public_key))
        database.commit()

        return attestation_hashes

    def _drop_attestation_table_data(self, attestation_hashes):
        """
        Remove all attestation data (claim based keys and ZKP blobs) by list of attestation hashes.

        :param attestation_hashes: hashes to remove
        :type attestation_hashes: [database_blob]
        :returns: None
        """
        if not attestation_hashes:
            return

        self.attestation_overlay.database.execute((u"DELETE FROM %s" % self.attestation_overlay.database.db_name)
                                                  + u" WHERE hash IN ("
                                                  + u", ".join(c for c in u"?" * len(attestation_hashes))
                                                  + u")",
                                                  attestation_hashes)
        self.attestation_overlay.database.commit()

    @docs(
        tags=["Attestation"],
        summary="Get information from the AttestationCommunity.",
        parameters=[{
            'in': 'query',
            'name': 'type',
            'description': 'Type of query',
            'type': 'string',
            'enum': ['drop_identity', 'outstanding', 'outstanding_verify', 'verification_output', 'peers', 'attributes'],
            'required': True
        }, {
            'in': 'query',
            'name': 'mid',
            'description': 'Filter by mid (only works for type=attributes)',
            'type': 'string'
        }],
        description="""
        type=drop_identity
        type=outstanding -> [(mid_b64, attribute_name)]
        type=outstanding_verify -> [(mid_b64, attribute_name)]
        type=verification_output -> {hash_b64: [(value_b64, match)]}
        type=peers -> [mid_b64]
        type=attributes&mid=mid_b64 -> [(attribute_name, attribute_hash)]
        """
    )
    async def handle_get(self, request):
        if not self.attestation_overlay or not self.identity_overlay:
            return Response({"error": "attestation or identity community not found"}, status=HTTP_NOT_FOUND)

        if not request.query or 'type' not in request.query:
            return Response({"error": "parameters or type missing"}, status=HTTP_BAD_REQUEST)

        if request.query['type'] == 'outstanding':
            formatted = []
            for k, v in self.attestation_requests.items():
                formatted.append(k + (v[1], ))
            return Response([(x, y, z) for x, y, z in formatted])

        elif request.query['type'] == 'outstanding_verify':
            formatted = self.verify_requests.keys()
            return Response([(x, y) for x, y in formatted])

        elif request.query['type'] == 'verification_output':
            formatted = {}
            for k, v in self.verification_output.items():
                formatted[b64encode(k).decode('utf-8')] = [(b64encode(a).decode('utf-8'), m) for a, m in v]
            return Response(formatted)

        elif request.query['type'] == 'peers':
            peers = self.session.network.get_peers_for_service(self.identity_overlay.community_id)
            return Response([b64encode(p.mid).decode('utf-8') for p in peers])

        elif request.query['type'] == 'attributes':
            if 'mid' in request.query:
                mid_b64 = request.query['mid']
                peer = self.get_peer_from_mid(mid_b64)
            else:
                peer = self.identity_overlay.my_peer
            if peer:
                pseudonym = self.identity_overlay.identity_manager.get_pseudonym(peer.public_key)
                trimmed = {}
                for credential in pseudonym.get_credentials():
                    # TODO: add support for more attesters
                    attestations = list(credential.attestations)
                    if attestations:
                        authority = self.identity_overlay.identity_manager.database.get_authority(attestations[0])
                        attester = b64encode(sha1(authority).digest()).decode()
                    else:
                        attester = ""
                    attribute_hash = pseudonym.tree.elements[credential.metadata.token_pointer].content_hash
                    json_metadata = json.loads(credential.metadata.serialized_json_dict)
                    trimmed[attribute_hash] = (json_metadata["name"], json_metadata, attester)
                # List of (name, attribute_hash, metadata, attester)
                return Response([(
                    data[0],
                    b64encode(strip_sha1_padding(attribute_hash)).decode(),
                    data[1],
                    data[2]) for attribute_hash, data in trimmed.items()])
            else:
                return Response([])

        elif request.query['type'] == 'drop_identity':
            to_keep = [self.persistent_key.public_key.key_to_bin()]
            if 'keep' in request.query:
                to_keep += [self.identity_overlay.my_peer.public_key.key_to_bin()]

            # Remove identity metadata and attestation proofing data, except for the keys to keep
            attestation_hashes = self._drop_identity_table_data(to_keep)
            self._drop_attestation_table_data(attestation_hashes)

            # Remove pending attestations
            self.attestation_requests.clear()

            # Generate new key
            my_new_peer = Peer(default_eccrypto.generate_key(u"curve25519"))
            identity_manager = self.identity_overlay.identity_manager
            await self.session.unload_overlay(self.identity_overlay)
            self.identity_overlay = await create_community(my_new_peer.key, self.session, identity_manager,
                                                           endpoint=self.session.endpoint)
            for overlay in self.session.overlays:
                overlay.my_peer = my_new_peer
            return Response({"success": True})

        else:
            return Response({"error": "type argument incorrect"}, status=HTTP_BAD_REQUEST)

    @docs(
        tags=["Attestation"],
        summary="Send a command to the AttestationCommunity.",
        parameters=[{
            'in': 'query',
            'name': 'type',
            'description': 'Type of query',
            'type': 'string',
            'enum': ['request', 'allow_verify', 'attest', 'verify'],
            'required': True
        }],
        description="""
        type=request&mid=mid_b64&attibute_name=attribute_name&id_format=id_format
        type=allow_verify&mid=mid_b64&attibute_name=attribute_name
        type=attest&mid=mid_b64&attribute_name=attribute_name&attribute_value=attribute_value_b64
        type=verify&mid=mid_b64&attribute_hash=attribute_hash_b64&id_format=id_format
                   &attribute_values=attribute_value_b64,...
        """
    )
    async def handle_post(self, request):
        if not self.attestation_overlay or not self.identity_overlay:
            return Response({"error": "attestation or identity community not found"}, status=HTTP_NOT_FOUND)

        args = request.query
        if not args or 'type' not in args:
            return Response({"error": "parameters or type missing"}, status=HTTP_BAD_REQUEST)

        if args['type'] == 'request':
            mid_b64 = args['mid']
            attribute_name = args['attribute_name']
            id_format = args.get('id_format', 'id_metadata')
            peer = self.get_peer_from_mid(mid_b64)
            if peer:
                key = self.attestation_overlay.get_id_algorithm(id_format).generate_secret_key()
                metadata = {"id_format": id_format}
                if 'metadata' in args:
                    metadata_unicode = json.loads(b64decode(args['metadata']))
                    metadata.update(metadata_unicode)
                self.attestation_metadata[(self.identity_overlay.my_peer, attribute_name)] = metadata
                self.attestation_overlay.request_attestation(peer, attribute_name, key, metadata)
                return Response({"success": True})
            else:
                return Response({"error": "peer unknown"}, status=HTTP_BAD_REQUEST)

        elif args['type'] == 'attest':
            mid_b64 = args['mid']
            attribute_name = args['attribute_name']
            attribute_value_b64 = args['attribute_value']
            outstanding = self.attestation_requests.pop((mid_b64, attribute_name))
            outstanding[0].set_result(b64decode(attribute_value_b64))
            return Response({"success": True})

        elif args['type'] == 'import_blob':
            # Import self-attested binary data
            attribute_name = args['attribute_name']
            id_format = args['id_format']
            metadata = {"id_format": id_format}
            if 'metadata' in args:
                metadata_unicode = json.loads(b64decode(args['metadata']))
                for k, v in metadata_unicode.items():
                    metadata[k.encode()] = v.encode()
            blob = await request.read()

            self.attestation_overlay.dump_blob(attribute_name, id_format, blob, metadata)

            return Response({"success": True})

        elif args['type'] == 'allow_verify':
            mid_b64 = args['mid']
            attribute_name = args['attribute_name']
            outstanding = self.verify_requests.pop((mid_b64, attribute_name))
            outstanding.set_result(True)
            return Response({"success": True})

        elif args['type'] == 'verify':
            mid_b64 = args['mid']
            attribute_hash = b64decode(args['attribute_hash'])
            reference_values = [b64decode(v) for v in args['attribute_values'].split(',')]
            id_format = args.get('id_format', 'id_metadata')
            peer = self.get_peer_from_mid(mid_b64)
            if peer:
                self.verification_output[attribute_hash] = \
                    [(b64decode(v), 0.0) for v in args['attribute_values'].split(',')]
                self.attestation_overlay.verify_attestation_values(peer.address, attribute_hash, reference_values,
                                                                   self.on_verification_results, id_format)
                return Response({"success": True})
            else:
                return Response({"error": "peer unknown"}, status=HTTP_BAD_REQUEST)

        else:
            return Response({"error": "type argument incorrect"}, status=HTTP_BAD_REQUEST)



import random
from binascii import unhexlify

from ..base import TestBase
from ...keyvault.crypto import default_eccrypto
from ...peer import Peer
from ...peerdiscovery.network import Network


def _generate_peer():
    key = default_eccrypto.generate_key(u'very-low')
    address = (".".join([str(random.randint(0, 255)) for _ in range(4)]), random.randint(0, 65535))
    return Peer(key, address)


class TestNetwork(TestBase):

    peers = [_generate_peer() for _ in range(4)]
    serialized_ipv4_len = 7

    def setUp(self):
        super(TestNetwork, self).setUp()
        self.network = Network()

    def test_discover_address(self):
        """
        Check registration of introducer and introduced when a new address is discovered.

        The introducer should be verified and not walkable.
        The introduced should not be verified and walkable.
        """
        self.network.discover_address(self.peers[0], self.peers[1].address)

        self.assertNotIn(self.peers[0].address, self.network.get_walkable_addresses())
        self.assertIn(self.peers[1].address, self.network.get_walkable_addresses())
        self.assertIn(self.peers[0], self.network.verified_peers)
        self.assertNotIn(self.peers[1], self.network.verified_peers)
        self.assertIn(self.peers[1].address, self.network.get_introductions_from(self.peers[0]))

    def test_discover_address_duplicate(self):
        """
        Check registration of introducer and introduced when the same address is discovered twice.
        """
        self.network.discover_address(self.peers[0], self.peers[1].address)
        self.network.discover_address(self.peers[0], self.peers[1].address)

        self.assertNotIn(self.peers[0].address, self.network.get_walkable_addresses())
        self.assertIn(self.peers[1].address, self.network.get_walkable_addresses())
        self.assertIn(self.peers[0], self.network.verified_peers)
        self.assertNotIn(self.peers[1], self.network.verified_peers)
        self.assertIn(self.peers[1].address, self.network.get_introductions_from(self.peers[0]))

    def test_discover_address_known(self):
        """
        Check if an address is already known, the network isn't updated.
        """
        self.network.discover_address(self.peers[0], self.peers[1].address)
        self.network.discover_address(self.peers[2], self.peers[1].address)

        self.assertNotIn(self.peers[0].address, self.network.get_walkable_addresses())
        self.assertIn(self.peers[1].address, self.network.get_walkable_addresses())
        self.assertNotIn(self.peers[2].address, self.network.get_walkable_addresses())
        self.assertIn(self.peers[0], self.network.verified_peers)
        self.assertNotIn(self.peers[1], self.network.verified_peers)
        self.assertIn(self.peers[2], self.network.verified_peers)
        self.assertIn(self.peers[1].address, self.network.get_introductions_from(self.peers[0]))

    def test_discover_address_known_parent_deceased(self):
        """
        Check if an address is already known, the new introducer adopts the introduced.
        """
        self.network.discover_address(self.peers[0], self.peers[1].address)
        self.network.remove_peer(self.peers[0])
        self.network.discover_address(self.peers[2], self.peers[1].address)

        self.assertIn(self.peers[1].address, self.network.get_walkable_addresses())
        self.assertNotIn(self.peers[2].address, self.network.get_walkable_addresses())
        self.assertIn(self.peers[2], self.network.verified_peers)
        self.assertNotIn(self.peers[1], self.network.verified_peers)
        self.assertIn(self.peers[1].address, self.network.get_introductions_from(self.peers[2]))

    def test_discover_address_blacklist(self):
        """
        Check if an address is in the blacklist, the network isn't updated.
        """
        self.network.blacklist.append(self.peers[2].address)
        self.network.discover_address(self.peers[0], self.peers[1].address)
        self.network.discover_address(self.peers[0], self.peers[2].address)

        self.assertNotIn(self.peers[0].address, self.network.get_walkable_addresses())
        self.assertIn(self.peers[1].address, self.network.get_walkable_addresses())
        self.assertNotIn(self.peers[2].address, self.network.get_walkable_addresses())
        self.assertIn(self.peers[0], self.network.verified_peers)
        self.assertNotIn(self.peers[1], self.network.verified_peers)
        self.assertNotIn(self.peers[2], self.network.verified_peers)
        self.assertIn(self.peers[1].address, self.network.get_introductions_from(self.peers[0]))

    def test_discover_address_multiple(self):
        """
        Check if a single peer can perform multiple introductions.
        """
        self.network.discover_address(self.peers[0], self.peers[1].address)
        self.network.discover_address(self.peers[0], self.peers[2].address)

        self.assertNotIn(self.peers[0].address, self.network.get_walkable_addresses())
        self.assertIn(self.peers[0], self.network.verified_peers)

        for other in [1, 2]:
            self.assertIn(self.peers[other].address, self.network.get_walkable_addresses())
            self.assertNotIn(self.peers[other], self.network.verified_peers)
            self.assertIn(self.peers[other].address, self.network.get_introductions_from(self.peers[0]))

    def test_get_introductions_from_cache_no_refresh(self):
        """
        Cache entries in the introduction cache should not refresh.

        This is to avoid dead peers sticking around the cache for too long.
        """
        intro1 = ('1.2.3.4', 5)
        intro2 = ('6.7.8.9', 10)
        intro3 = ('11.12.13.14', 15)
        self.network.discover_address(self.peers[0], intro1)
        self.network.discover_address(self.peers[1], intro2)

        ilist0 = self.network.get_introductions_from(self.peers[0])
        ilist1 = self.network.get_introductions_from(self.peers[1])
        self.assertListEqual([self.peers[0], self.peers[1]], list(self.network.reverse_intro_lookup))
        self.assertListEqual([intro1], ilist0)
        self.assertListEqual([intro2], ilist1)

        # We should make a good faith attempt to update the cache with a new introduction.
        self.network.discover_address(self.peers[0], intro3)
        ilist2 = self.network.get_introductions_from(self.peers[0])
        self.assertListEqual([self.peers[0], self.peers[1]], list(self.network.reverse_intro_lookup))
        self.assertListEqual([intro1, intro3], ilist2)

    def test_discover_services(self):
        """
        Check if services are properly registered for a peer.
        """
        service = bytes(range(20))
        self.network.discover_services(self.peers[0], [service])
        self.network.add_verified_peer(self.peers[0])

        self.assertIn(service, self.network.get_services_for_peer(self.peers[0]))
        self.assertIn(self.peers[0], self.network.get_peers_for_service(service))

    def test_discover_services_unverified(self):
        """
        Check if services are properly registered for an unverified peer.
        """
        service = bytes(range(20))
        self.network.discover_services(self.peers[0], [service])

        self.assertIn(service, self.network.get_services_for_peer(self.peers[0]))
        self.assertNotIn(self.peers[0], self.network.get_peers_for_service(service))

    def test_discover_services_update(self):
        """
        Check if services are properly combined for a peer.
        """
        service1 = bytes(range(20))
        service2 = bytes(range(20, 40))
        self.network.discover_services(self.peers[0], [service1])
        self.network.discover_services(self.peers[0], [service2])
        self.network.add_verified_peer(self.peers[0])

        self.assertIn(service1, self.network.get_services_for_peer(self.peers[0]))
        self.assertIn(service2, self.network.get_services_for_peer(self.peers[0]))
        self.assertIn(self.peers[0], self.network.get_peers_for_service(service1))
        self.assertIn(self.peers[0], self.network.get_peers_for_service(service2))

    def test_discover_services_update_overlap(self):
        """
        Check if services are properly combined when discovered services overlap.
        """
        service1 = bytes(range(20))
        service2 = bytes(range(20, 40))
        self.network.discover_services(self.peers[0], [service1])
        self.network.discover_services(self.peers[0], [service1, service2])
        self.network.add_verified_peer(self.peers[0])

        self.assertIn(service1, self.network.get_services_for_peer(self.peers[0]))
        self.assertIn(service2, self.network.get_services_for_peer(self.peers[0]))
        self.assertIn(self.peers[0], self.network.get_peers_for_service(service1))
        self.assertIn(self.peers[0], self.network.get_peers_for_service(service2))

    def test_discover_services_address_update(self):
        """
        Check if an address update gets processed correctly.
        """
        service1 = bytes(range(20))
        service2 = bytes(range(20, 40))

        pk = self.peers[0].public_key
        peer = Peer(pk, ('1.1.1.1', 1))
        self.network.add_verified_peer(peer)
        self.network.discover_services(peer, [service1])

        peer = Peer(pk, ('1.1.1.1', 1))
        self.network.add_verified_peer(peer)
        self.network.discover_services(peer, [service2])

        peer = Peer(pk, ('1.1.1.1', 2))
        self.network.add_verified_peer(peer)
        self.network.discover_services(peer, [service1])

        peer = Peer(pk, ('1.1.1.1', 2))
        self.network.add_verified_peer(peer)
        self.network.discover_services(peer, [service2])

        self.assertEqual(self.network.get_peers_for_service(service1)[0].address, peer.address)
        self.assertEqual(self.network.get_peers_for_service(service2)[0].address, peer.address)

    def test_discover_services_cache(self):
        """
        Check if services cache updates properly.
        """
        service1 = bytes(range(20))
        service2 = bytes(range(20, 40))
        self.network.reverse_service_cache_size = 1
        self.network.discover_services(self.peers[0], [service1])
        self.network.add_verified_peer(self.peers[0])
        self.network.discover_services(self.peers[1], [service2])
        self.network.add_verified_peer(self.peers[1])

        self.network.get_peers_for_service(service1)
        self.network.get_peers_for_service(service2)

        self.assertListEqual([service2], list(self.network.reverse_service_lookup))

    def test_add_verified_peer_new(self):
        """
        Check if a new verified peer can be added to the network.
        """
        self.network.add_verified_peer(self.peers[0])

        self.assertNotIn(self.peers[0].address, self.network.get_walkable_addresses())
        self.assertIn(self.peers[0], self.network.verified_peers)
        self.assertListEqual([], self.network.get_introductions_from(self.peers[0]))

    def test_add_verified_peer_blacklist(self):
        """
        Check if a new verified peer can be added to the network.
        """
        self.network.blacklist.append(self.peers[0].address)
        self.network.add_verified_peer(self.peers[0])

        self.assertNotIn(self.peers[0].address, self.network.get_walkable_addresses())
        self.assertNotIn(self.peers[0], self.network.verified_peers)

    def test_add_verified_peer_duplicate(self):
        """
        Check if an already verified (by slightly changed) peer doesn't cause duplicates in the network.
        """
        self.network.add_verified_peer(self.peers[0])
        self.peers[0].update_clock(1)
        self.network.add_verified_peer(self.peers[0])

        self.assertNotIn(self.peers[0].address, self.network.get_walkable_addresses())
        self.assertIn(self.peers[0], self.network.verified_peers)
        self.assertListEqual([], self.network.get_introductions_from(self.peers[0]))

    def test_add_verified_peer_promote(self):
        """
        Check if a peer can be promoted from an address to a verified peer.
        """
        self.network.discover_address(self.peers[1], self.peers[0].address)
        self.network.add_verified_peer(self.peers[0])

        self.assertNotIn(self.peers[0].address, self.network.get_walkable_addresses())
        self.assertIn(self.peers[0], self.network.verified_peers)
        self.assertListEqual([], self.network.get_introductions_from(self.peers[0]))

    def test_get_verified_by_address(self):
        """
        Check if we can find a peer in our network by its address.
        """
        self.network.add_verified_peer(self.peers[0])

        self.assertEqual(self.peers[0], self.network.get_verified_by_address(self.peers[0].address))
        self.assertIn(self.peers[0].address, self.network.reverse_ip_lookup)

    def test_get_verified_by_address_cache_pop(self):
        """
        When the cache if full, pop the least-used entry.
        """
        self.network.add_verified_peer(self.peers[0])
        self.network.add_verified_peer(self.peers[1])
        self.network.reverse_ip_cache_size = 1

        self.network.get_verified_by_address(self.peers[1].address)  # Cache stack: [1] (full)
        self.assertListEqual([self.peers[1].address], list(self.network.reverse_ip_lookup))

        self.network.get_verified_by_address(self.peers[0].address)  # Cache stack: [0] (full)
        self.assertListEqual([self.peers[0].address], list(self.network.reverse_ip_lookup))

    def test_get_verified_by_address_cache_refresh(self):
        """
        Asking for the same peer twice should land it back on top of the cleanup stack.
        """
        self.network.add_verified_peer(self.peers[0])
        self.network.add_verified_peer(self.peers[1])
        self.network.reverse_ip_cache_size = 2

        self.network.get_verified_by_address(self.peers[1].address)  # Cache stack: [1]
        self.network.get_verified_by_address(self.peers[0].address)  # Cache stack: [1, 0] (full)
        self.network.get_verified_by_address(self.peers[1].address)  # Cache stack: [0, 1] (full)

        self.assertListEqual([self.peers[0].address, self.peers[1].address], list(self.network.reverse_ip_lookup))

    def test_get_verified_by_public_key(self):
        """
        Check if we can find a peer in our network by its public key.
        """
        self.network.add_verified_peer(self.peers[0])

        self.assertEqual(self.peers[0],
                         self.network.get_verified_by_public_key_bin(self.peers[0].public_key.key_to_bin()))

    def test_remove_by_address(self):
        """
        Check if we can remove a peer from our network by its address.
        """
        self.network.add_verified_peer(self.peers[0])
        self.network.discover_services(self.peers[0], [b"0" * 20])
        self.network.remove_by_address(self.peers[0].address)

        self.assertNotIn(self.peers[0], self.network.verified_peers)
        self.assertNotIn(self.peers[0].address, self.network.get_walkable_addresses())
        self.assertEqual(set(), self.network.get_services_for_peer(self.peers[0]))

    def test_remove_by_address_unverified(self):
        """
        Check if we can remove an unverified peer from our network by its address.
        """
        self.network.discover_address(self.peers[0], self.peers[1].address)
        self.network.remove_by_address(self.peers[1].address)

        self.assertNotIn(self.peers[1].address, self.network.get_walkable_addresses())

    def test_remove_by_address_unknown(self):
        """
        Removing unknown peers should not affect other peers.
        """
        self.network.add_verified_peer(self.peers[0])

        previous_walkable = self.network.get_walkable_addresses()
        previous_verified = self.network.verified_peers

        self.network.remove_by_address(self.peers[1].address)

        self.assertEqual(previous_walkable, self.network.get_walkable_addresses())
        self.assertEqual(previous_verified, self.network.verified_peers)

    def test_remove_by_address_no_services(self):
        """
        Check if we can remove a peer from our network if it doesn't have services by address.
        """
        self.network.add_verified_peer(self.peers[0])
        self.network.remove_by_address(self.peers[0].address)

        self.assertNotIn(self.peers[0], self.network.verified_peers)
        self.assertNotIn(self.peers[0].address, self.network.get_walkable_addresses())

    def test_remove_peer(self):
        """
        Check if we can remove a peer from our network.
        """
        self.network.add_verified_peer(self.peers[0])
        self.network.discover_services(self.peers[0], [b"0" * 20])
        self.network.remove_peer(self.peers[0])

        self.assertNotIn(self.peers[0], self.network.verified_peers)
        self.assertNotIn(self.peers[0].address, self.network.get_walkable_addresses())
        self.assertEqual(set(), self.network.get_services_for_peer(self.peers[0]))

    def test_remove_peer_external(self):
        """
        Check if we can remove an externally created peer from our network.
        """
        self.network.discover_address(self.peers[1], self.peers[0].address)
        self.network.remove_peer(self.peers[0])

        self.assertNotIn(self.peers[0].address, self.network.get_walkable_addresses())

    def test_remove_peer_unknown(self):
        """
        Removing unknown peers should not affect other peers.
        """
        self.network.add_verified_peer(self.peers[0])

        previous_walkable = self.network.get_walkable_addresses()
        previous_verified = self.network.verified_peers

        self.network.remove_peer(self.peers[1])

        self.assertEqual(previous_walkable, self.network.get_walkable_addresses())
        self.assertEqual(previous_verified, self.network.verified_peers)

    def test_get_walkable_by_service(self):
        """
        Check if we can retrieve walkable addresses by parent service id.
        """
        service = bytes(range(20))
        self.network.discover_address(self.peers[2], self.peers[3].address)
        self.network.discover_address(self.peers[0], self.peers[1].address)
        self.network.discover_services(self.peers[0], [service])

        self.assertEqual([self.peers[1].address], self.network.get_walkable_addresses(service))

    def test_snapshot_only_verified(self):
        """
        Check if a snapshot poperly serializes verified peers.
        """
        for peer in self.peers:
            self.network.add_verified_peer(peer)
        snapshot = self.network.snapshot()

        self.assertEqual(len(snapshot), TestNetwork.serialized_ipv4_len * len(self.peers))

    def test_snapshot_verified_and_unverified(self):
        """
        Check if a snapshot poperly serializes only verified peers.
        """
        self.network.add_verified_peer(self.peers[0])
        self.network.discover_address(self.peers[0], self.peers[1].address)
        snapshot = self.network.snapshot()

        self.assertEqual(len(snapshot), TestNetwork.serialized_ipv4_len)

    def test_snapshot_only_unverified(self):
        """
        Check if a snapshot is empty without verified peers.
        """
        self.network.blacklist_mids.append(self.peers[0].mid)
        for peer in self.peers[1:]:
            self.network.discover_address(self.peers[0], peer.address)
        snapshot = self.network.snapshot()

        self.assertEqual(len(snapshot), 0)

    def test_snapshot_no_peers(self):
        """
        Check if a snapshot is empty without verified peers.
        """
        snapshot = self.network.snapshot()

        self.assertEqual(len(snapshot), 0)

    def test_load_snapshot(self):
        """
        Check if peers can be properly loaded from a snapshot.
        """
        expected = {
            ('54.226.56.113', 57674),
            ('70.200.183.29', 60815),
            ('194.116.42.174', 21882),
            ('143.93.254.175', 41862)
        }
        snapshot = unhexlify("01" "36e23871" "e14a"
                             "01" "46c8b71d" "ed8f"
                             "01" "c2742aae" "557a"
                             "01" "8f5dfeaf" "a386")

        self.network.load_snapshot(snapshot)
        peers = set(self.network.get_walkable_addresses())

        self.assertSetEqual(peers, expected)

    def test_load_snapshot_empty(self):
        """
        Check if no peers are loaded from an empty snapshot.
        """
        self.network.load_snapshot("")
        peers = set(self.network.get_walkable_addresses())

        self.assertSetEqual(peers, set())

    def test_load_snapshot_malformed(self):
        """
        Check if no peers are loaded from a malformed snapshot.
        """
        snapshot = unhexlify("0136e23871e14a0146c8b71ded8f01c2742aae557a018f5dfeafa386")

        self.network.load_snapshot(snapshot[1:])
        peers = set(self.network.get_walkable_addresses())

        self.assertSetEqual(peers, set())

    def test_load_snapshot_truncated(self):
        """
        Check if as many peers as possible are loaded from a truncated snapshot.
        """
        snapshot = unhexlify("0136e23871e14a0146c8b71ded8f01c2742aae557a018f5dfeafa386")

        self.network.load_snapshot(snapshot[:-1])
        peers = set(self.network.get_walkable_addresses())

        self.assertEqual(3, len(peers))



from binascii import unhexlify
from unittest.mock import Mock

from .test_datachecker import dht_pkt, ipv8_pkt, tracker_pkt, tunnel_pkt, utp_pkt
from ...base import TestBase
from ....messaging.anonymization.tunnel import PEER_FLAG_EXIT_BT, PEER_FLAG_EXIT_IPV8, TunnelExitSocket


class TestExitSocket(TestBase):

    async def test_is_allowed(self):
        """
        Check if the ExitSocket correctly detects forbidden packets.
        """
        overlay = Mock(_prefix=unhexlify('000281ded07332bdc775aa5a46f96de9f8f390bbc9f3'))
        exit_socket = TunnelExitSocket(0, None, overlay)

        overlay.settings.peer_flags = {}
        self.assertFalse(exit_socket.is_allowed(tracker_pkt))
        self.assertFalse(exit_socket.is_allowed(dht_pkt))
        self.assertFalse(exit_socket.is_allowed(utp_pkt))
        self.assertFalse(exit_socket.is_allowed(ipv8_pkt))
        self.assertTrue(exit_socket.is_allowed(tunnel_pkt))

        overlay.settings.peer_flags = {PEER_FLAG_EXIT_BT}
        self.assertTrue(exit_socket.is_allowed(tracker_pkt))
        self.assertTrue(exit_socket.is_allowed(dht_pkt))
        self.assertTrue(exit_socket.is_allowed(utp_pkt))
        self.assertFalse(exit_socket.is_allowed(ipv8_pkt))
        self.assertTrue(exit_socket.is_allowed(tunnel_pkt))

        overlay.settings.peer_flags = {PEER_FLAG_EXIT_IPV8}
        self.assertFalse(exit_socket.is_allowed(tracker_pkt))
        self.assertFalse(exit_socket.is_allowed(dht_pkt))
        self.assertFalse(exit_socket.is_allowed(utp_pkt))
        self.assertTrue(exit_socket.is_allowed(ipv8_pkt))
        self.assertTrue(exit_socket.is_allowed(tunnel_pkt))

        overlay.settings.peer_flags = {PEER_FLAG_EXIT_BT, PEER_FLAG_EXIT_IPV8}
        self.assertTrue(exit_socket.is_allowed(tracker_pkt))
        self.assertTrue(exit_socket.is_allowed(dht_pkt))
        self.assertTrue(exit_socket.is_allowed(utp_pkt))
        self.assertTrue(exit_socket.is_allowed(ipv8_pkt))
        self.assertTrue(exit_socket.is_allowed(tunnel_pkt))

        await exit_socket.close()



from collections import defaultdict

# Map of info_hash -> peer list
global_dht_services = defaultdict(list)


class MockDHTProvider:

    def __init__(self, peer):
        self.peer = peer
        # DHTDiscoveryCommunity functionality
        global_dht_services[peer.mid].append(peer)

    async def peer_lookup(self, mid, peer=None):  # pylint: disable=W0613
        return await self.lookup(mid)

    async def lookup(self, info_hash):
        return info_hash, global_dht_services.get(info_hash, [])

    async def announce(self, info_hash, intro_point):
        global_dht_services[info_hash].append(intro_point)



import logging
import os
import shutil
import sys
import threading
import time
import uuid
from asyncio import all_tasks, ensure_future, get_event_loop, iscoroutine, sleep
from functools import partial

import asynctest

from .mocking.endpoint import internet
from .mocking.ipv8 import MockIPv8
from ..peer import Peer

try:
    get_event_loop().set_debug(True)
except RuntimeError:
    logging.warning("Failed to set debug mode on the main event loop! "
                    "You may be missing out on asyncio output!")


def _on_packet_fragile_cb(self, packet, warn_unknown=True):
    """
    A fragile version of on_packet that crashes on message handling failures.

    These failures won't actually cause IPv8 to crash in production, but you should probably handle these.

    Add overlay classes to use in production mode to the ``production_overlay_classes`` list.
    Filter nodes to run in production mode by overwriting ``TestBase.patch_overlays``.
    """
    result = self.decode_map[packet[1][22]](*packet)
    if iscoroutine(result):
        self.register_anonymous_task('on_packet', ensure_future(result))


class TestBase(asynctest.TestCase):

    __testing__ = True
    __lockup_timestamp__ = 0

    # The time after which the whole test suite is os.exited
    MAX_TEST_TIME = 10

    def __init__(self, methodName='runTest'):
        super(TestBase, self).__init__(methodName)
        self.nodes = []
        self.overlay_class = object
        internet.clear()
        self._tempdirs = []
        self.production_overlay_classes = []
        self._uncaught_async_failure = None

    def initialize(self, overlay_class, node_count, *args, **kwargs):
        self.overlay_class = overlay_class
        self.nodes = [self.create_node(*args, **kwargs) for _ in range(node_count)]

        # Add nodes to each other.
        for node in self.nodes:
            for other in self.nodes:
                if other == node:
                    continue
                private_peer = other.my_peer
                public_peer = Peer(private_peer.public_key, private_peer.address)
                node.network.add_verified_peer(public_peer)
                node.network.discover_services(public_peer, [overlay_class.community_id])

        # Make packet handling fragile.
        for i in range(len(self.nodes)):
            self.patch_overlays(i)

    def _patch_overlay(self, overlay):
        if overlay and overlay.__class__ not in self.production_overlay_classes:
            overlay.on_packet = partial(_on_packet_fragile_cb, overlay)

    def patch_overlays(self, i):
        """
        Method to make the packet handlers of a particular node fragile.

        If you want to disable fragile packet handling for an entire class over overlays, add the class to the
        production_overlay_classes list.

        If you want to disable fragile packet handling for a particular node, overwrite this method.
        """
        self._patch_overlay(self.node(i).overlay)
        self._patch_overlay(self.node(i).dht)
        for overlay in self.node(i).overlays:
            self._patch_overlay(overlay)

    def _cb_exception(self, loop, context):
        """
        Callback for asyncio exceptions.

        Do not call `self.fail` or `pytest.fail` or any other failure mechanism in here.
        These methods work by raising an Exception, which will silently fail as this is not the main loop.
        """
        # Only fail on the first failure, to not cause exception shadowing on compound errors.
        if self._uncaught_async_failure is None:
            self._uncaught_async_failure = context

    def setUp(self):
        self.loop.set_debug(True)
        self.loop.set_exception_handler(self._cb_exception)
        super(TestBase, self).setUp()
        TestBase.__lockup_timestamp__ = time.time()

    async def tearDown(self):
        try:
            for node in self.nodes:
                await node.stop()
            internet.clear()
        finally:
            while self._tempdirs:
                shutil.rmtree(self._tempdirs.pop(), ignore_errors=True)
        # Now that everyone has calmed down, sweep up the remaining callbacks and check if they failed.
        await asynctest.helpers.exhaust_callbacks(self.loop)
        if self._uncaught_async_failure is not None:
            raise self._uncaught_async_failure["exception"]
        self.loop.set_exception_handler(None)  # None is equivalent to the default handler
        super(TestBase, self).tearDown()

    @classmethod
    def setUpClass(cls):
        TestBase.__lockup_timestamp__ = time.time()

        # pytest has its own timeout.
        if "PYTEST_CURRENT_TEST" not in os.environ:
            def check_loop():
                while time.time() - TestBase.__lockup_timestamp__ < cls.MAX_TEST_TIME:
                    time.sleep(2)
                    # If the test class completed normally, exit
                    if not cls.__testing__:
                        return
                # If we made it here, there is a serious issue which we cannot recover from.
                # Most likely the threadpool got into a deadlock while shutting down.
                import traceback
                print("The test-suite locked up! Force quitting! Thread dump:", file=sys.stderr)  # noqa: T001
                for tid, stack in sys._current_frames().items():
                    if tid != threading.currentThread().ident:
                        print("THREAD#%d" % tid, file=sys.stderr)  # noqa: T001
                        for line in traceback.format_list(traceback.extract_stack(stack)):
                            print("|", line[:-1].replace('\n', '\n|   '), file=sys.stderr)  # noqa: T001

                tasks = all_tasks(get_event_loop())
                if tasks:
                    print("Pending tasks:")  # noqa: T001
                    for task in tasks:
                        print(">     %s" % task)  # noqa: T001

                # Our test suite catches the SIGINT signal, this allows it to print information before force exiting.
                # If we were to hard exit here (through os._exit) we would lose this additional information.
                import signal
                os.kill(os.getpid(), signal.SIGINT)
                # But sometimes it just flat out refuses to die (sys.exit will also not work in this case).
                # So we double kill ourselves:
                time.sleep(5.0)  # Just in case anyone is listening to our signal and wishes to log some stats quickly.
                os._exit(1)  # pylint: disable=W0212
            t = threading.Thread(target=check_loop)
            t.daemon = True
            t.start()

    @classmethod
    def tearDownClass(cls):
        cls.__testing__ = False

    def create_node(self, *args, **kwargs):
        return MockIPv8(u"low", self.overlay_class, *args, **kwargs)

    def add_node_to_experiment(self, node):
        """
        Add a new node to this experiment (use `create_node()`).
        """
        for other in self.nodes:
            private_peer = other.my_peer
            public_peer = Peer(private_peer.public_key, private_peer.address)
            node.network.add_verified_peer(public_peer)
            node.network.discover_services(public_peer, node.overlay.community_id)
        self.nodes.append(node)

    @staticmethod
    def is_background_task(task):
        # Only in Python 3.8+ will we have a get_name function
        name = task.get_name() if hasattr(task, 'get_name') else getattr(task, 'name', f'Task-{id(task)}')
        return name.endswith('_check_tasks')

    async def deliver_messages(self, timeout=.1):
        """
        Allow peers to communicate.

        The strategy is as follows:
         1. Measure the amount of existing asyncio tasks
         2. After 10 milliseconds, check if we are below 2 tasks twice in a row
         3. If not, go back to handling calls (step 2) or return, if the timeout has been reached

        :param timeout: the maximum time to wait for messages to be delivered
        """
        rtime = 0
        probable_exit = False

        while rtime < timeout:
            await sleep(.01)
            rtime += .01
            if len([task for task in all_tasks() if not self.is_background_task(task)]) < 2:
                if probable_exit:
                    break
                probable_exit = True
            else:
                probable_exit = False

    async def introduce_nodes(self):
        for node in self.nodes:
            for other in self.nodes:
                if other != node:
                    node.overlay.walk_to(other.endpoint.wan_address)
        await self.deliver_messages()

    def temporary_directory(self):
        rndstr = '_temp_' + uuid.uuid4().hex
        d = os.path.abspath(self.__class__.__name__ + rndstr)
        self._tempdirs.append(d)
        os.makedirs(d)
        return d

    def address(self, i):
        return self.peer(i).address

    def endpoint(self, i):
        return self.nodes[i].endpoint

    def key_bin(self, i):
        return self.nodes[i].my_peer.public_key.key_to_bin()

    def key_bin_private(self, i):
        return self.nodes[i].my_peer.key.key_to_bin()

    def mid(self, i):
        return self.nodes[i].my_peer.mid

    def my_peer(self, i):
        return self.nodes[i].overlay.my_peer

    def network(self, i):
        return self.nodes[i].network

    def node(self, i):
        return self.nodes[i]

    def overlay(self, i):
        return self.nodes[i].overlay

    def peer(self, i):
        return Peer(self.nodes[i].my_peer.public_key.key_to_bin(), self.nodes[i].endpoint.wan_address)

    def private_key(self, i):
        return self.nodes[i].my_peer.key

    def public_key(self, i):
        return self.nodes[i].my_peer.public_key
